# Copyright (C) 2021-2024 Phoebe Pearce
#
# This file is part of RayFlare and is released under the GNU Lesser General Public License (LGPL), version 3.
# Please see the LICENSE.txt file included as part of this package.
#
# Contact: p.pearce@unsw.edu.au

import xarray as xr
import numpy as np
from rayflare.transfer_matrix_method.tmm import tmm_structure
import os
from solcore.absorption_calculator import OptiStack

import logging
logging.basicConfig(format='%(levelname)s: %(message)s',level=logging.INFO)

def make_TMM_lookuptable(
    layers,
    incidence,
    transmission,
    surf_name,
    options,
    structpath,
    coherent=True,
    coherency_list=None,
    prof_layers=None,
    sides=None,
    overwrite=False,
):
    """
    Takes a layer stack and calculates and stores lookup tables for use with the ray-tracer.

    :param layers: a list of layers. These can be Solcore 'Layer' objects, or any other layer format accepted \
    by the Solcore class 'OptiStack'.
    :param incidence: semi-incidence medium. Should be an instance of a Solcore material object
    :param transmission: semi-infinite transmission medium. Should be an instance of a Solcore material object
    :param surf_name: name of the surfaces, for storing the lookup table (string).
    :param options: dictionary or State object containing user options
    :param structpath: file path where matrices will be stored or loaded from
    :param coherent: boolean. True if all the layers in the stack (excluding the semi-infinite incidence and \
    transmission medium) are coherent, False otherwise. Default True.
    :param coherency_list: list. List of 'c' (coherent) and 'i' (incoherent) for each layer excluding incidence and \
    transmission media. Only needs to be provided if coherent = False. Default = None
    :param prof_layers: Indices of the layers in which the parameters relating to the absorption profile should be \
    calculated and stored. Layer 0 is the incidence medium.
    :param sides: List of which sides of incidence should all parameters be calculated for; 1 indicates incidence from \
    the front and -1 is rear incidence. Default = [1, -1]
    :param overwrite: boolean. If True, existing saved lookup tables will be overwritten. Default = False.
    :return: xarray Dataset with the R, A, T and (if relevant) absorption profile coefficients for each \
    wavelength, angle, polarization, side of incidence.
    """

    if sides is None:
        sides = [1, -1]

    savepath = os.path.join(structpath, surf_name + ".nc")
    if os.path.isfile(savepath) and not overwrite:
        logging.info("Existing lookup table found")
        allres = xr.open_dataset(savepath)
    else:
        wavelengths = options["wavelength"]
        n_angles = options["lookuptable_angles"]
        thetas = np.linspace(0, (np.pi / 2) - 1e-3, n_angles)
        if prof_layers is not None:
            profile = True
            prof_layers_rev = len(layers) - np.array(prof_layers[::-1]) + 1
            prof_layer_list = [prof_layers, prof_layers_rev.tolist()]
        else:
            profile = False
            prof_layer_list = [None, None]

        n_layers = len(layers)
        optlayers = OptiStack(layers, substrate=transmission, incidence=incidence)
        optlayers_flip = OptiStack(
            layers[::-1], substrate=incidence, incidence=transmission
        )
        optstacks = [optlayers, optlayers_flip]

        if coherency_list is not None:
            coherency_list = np.array(coherency_list)
            coherency_list[np.array(optlayers.widths) == 0] = "c" # incoherent implementation doesn't work/
            # make sense for zero thickness, coherent implementation will correctly ignore the layer
            coherency_lists = [coherency_list.tolist(), coherency_list[::-1].tolist()]
        else:
            coherency_lists = [["c"] * n_layers] * 2
        # can calculate by angle, already vectorized over wavelength
        pols = ["s", "p"]

        R = xr.DataArray(
            np.empty((2, 2, len(wavelengths), n_angles)),
            dims=["side", "pol", "wl", "angle"],
            coords={
                "side": sides,
                "pol": pols,
                "wl": wavelengths * 1e9,
                "angle": thetas,
            },
            name="R",
        )
        T = xr.DataArray(
            np.empty((2, 2, len(wavelengths), n_angles)),
            dims=["side", "pol", "wl", "angle"],
            coords={
                "side": sides,
                "pol": pols,
                "wl": wavelengths * 1e9,
                "angle": thetas,
            },
            name="T",
        )
        Alayer = xr.DataArray(
            np.empty((2, 2, n_angles, len(wavelengths), n_layers)),
            dims=["side", "pol", "angle", "wl", "layer"],
            coords={
                "side": sides,
                "pol": pols,
                "wl": wavelengths * 1e9,
                "angle": thetas,
                "layer": range(1, n_layers + 1),
            },
            name="Alayer",
        )

        if profile:
            Aprof = xr.DataArray(
                np.empty((2, 2, n_angles, 6, len(prof_layers), len(wavelengths))),
                dims=["side", "pol", "angle", "coeff", "layer", "wl"],
                coords={
                    "side": sides,
                    "pol": pols,
                    "wl": wavelengths * 1e9,
                    "angle": thetas,
                    "layer": prof_layers,
                    "coeff": ["A1", "A2", "A3_r", "A3_i", "a1", "a3"],
                },
                name="Aprof",
            )

        pass_options = {}

        pass_options["wavelength"] = wavelengths
        pass_options["depth_spacing"] = 1e5
        # we don't actually want to calculate a profile, so the depth spacing
        # doesn't matter, but it needs to be set to something. Larger value means we don't make extremely large arrays
        # no reason during the calculation

        for i1, side in enumerate(sides):
            prof_layer_side = prof_layer_list[i1]
            R_loop = np.empty((len(wavelengths), n_angles))
            T_loop = np.empty((len(wavelengths), n_angles))
            Alayer_loop = np.empty((n_angles, len(wavelengths), n_layers))
            if profile:
                Aprof_loop = np.empty((n_angles, 6, len(prof_layers), len(wavelengths)))

            pass_options["coherent"] = coherent
            pass_options["coherency_list"] = coherency_lists[i1]

            for pol in pols:

                for i3, theta in enumerate(thetas):

                    pass_options["pol"] = pol
                    pass_options["theta_in"] = theta

                    tmm_struct = tmm_structure(optstacks[i1])
                    res = tmm_struct.calculate(
                        pass_options, profile=profile, layers=prof_layer_side
                    )

                    R_loop[:, i3] = np.real(res["R"])
                    T_loop[:, i3] = np.real(res["T"])
                    Alayer_loop[i3, :, :] = np.real(res["A_per_layer"])

                    if profile:
                        Aprof_loop[i3, :, :, :] = np.real(res["profile_coeff"])

                # sometimes get very small negative values (like -1e-20)
                R_loop[R_loop < 0] = 0
                T_loop[T_loop < 0] = 0
                Alayer_loop[Alayer_loop < 0] = 0

                if side == -1:
                    Alayer_loop = np.flip(Alayer_loop, axis=2)
                    # layers were upside down to do calculation; want labelling to be with
                    # respect to side = 1 for consistency
                    if profile:
                        Aprof_loop = np.flip(Aprof_loop, axis=2)

                R.loc[dict(side=side, pol=pol)] = R_loop
                T.loc[dict(side=side, pol=pol)] = T_loop
                Alayer.loc[dict(side=side, pol=pol)] = Alayer_loop

                if profile:
                    Aprof.loc[dict(side=side, pol=pol)] = Aprof_loop
                    Aprof.transpose("side", "pol", "wl", "angle", "layer", "coeff")

        Alayer = Alayer.transpose("side", "pol", "wl", "angle", "layer")

        if profile:
            allres = xr.merge([R, T, Alayer, Aprof])
        else:
            allres = xr.merge([R, T, Alayer])

        unpol = allres.reduce(np.mean, "pol").assign_coords(pol="u").expand_dims("pol")
        allres = allres.merge(unpol)

        allres.to_netcdf(savepath)

    return allres



# Copyright (C) 2021-2024 Phoebe Pearce
#
# This file is part of RayFlare and is released under the GNU Lesser General Public License (LGPL), version 3.
# Please see the LICENSE.txt file included as part of this package.
#
# Contact: p.pearce@unsw.edu.au

import numpy as np
from sparse import load_npz, COO, stack, einsum
from rayflare.angles import make_angle_vector, fold_phi, overall_bin
import os
import xarray as xr
from solcore.state import State

from rayflare.structure import Interface, BulkLayer
from rayflare.utilities import get_savepath

from rayflare import logger

def calculate_RAT(SC, options, save_location="default"):
    """
    After the list of Interface and BulkLayers has been processed by process_structure,
    this function calculates the R, A and T by calling matrix_multiplication.

    :param SC: list of Interface and BulkLayer objects. Order is [Interface, BulkLayer, Interface]
    :param options: options for the matrix calculations (State object or dictionary)
    :param save_location: location from which to load the redistribution matrices. Current options:

              - 'default', which stores the results in folder in your home directory called 'RayFlare_results'
              - 'current', which stores the results in the current working directory
              - or you can specify the full path location for wherever you want the results to be stored.

            This should match what was specified for process_structure.

    :return: The number of returned values depends on whether absorption profiles were calculated or not. The first two
            are always returned, the final two are only returned if a calculation of absorption profiles was done.

            - RAT - an xarray with coordinates bulk_index and wl (wavelength), and 3 data variables: R (reflection),
              T (transmission) and A_bulk (absorption in the bulk medium. Currently, the bulk index can only be 0.
            - results_per_pass - a dictionary with entries 'r' (reflection), 't' (transmission), 'a' (absorption in the
              surfaces) and 'A' (bulk absorption), which store these quantities per each pass of the bulk or interaction
              with the relevant surface during matrix multiplication. 'r', 't' and 'A' are lists of length 1, corresponding
              to one set of values for each bulk material; the list entry is an array which is indexed as
              (pass number, wavelength). 'a' is a list of length two, corresponding to absorption in the front and back
              interface respectively. Each entry in the list is an array indexed as (pass number, wavelength, layer index).
            - profile - a list of xarrays, one for each surface. These store the absorption profiles and have coordinates
              wavelength and z (depth) position.
            - bulk_profile - a list of arrays, one for each bulk (currently, always only one). Indices are
              (wavelength, position)

    """

    if isinstance(options, dict):
        options = State(options)

    bulk_mats = []
    bulk_widths = []
    layer_widths = []
    layer_names = []
    calc_prof_list = []

    for struct in SC:
        if isinstance(struct, BulkLayer):
            bulk_mats.append(struct.material)
            bulk_widths.append(struct.width)

        if isinstance(struct, Interface):
            layer_names.append(struct.name)
            layer_widths.append((np.array(struct.widths) * 1e9).tolist())
            calc_prof_list.append(struct.prof_layers)

    results = matrix_multiplication(
        bulk_mats, bulk_widths, options, layer_names, calc_prof_list, save_location
    )

    return results


def make_v0(
    th_in, phi_in, num_wl, n_theta_bins, c_azimuth, phi_sym, theta_spacing="sin"
):
    """
    This function makes the v0 array, corresponding to the input power per angular channel
    at each wavelength, of size (num_wl, n_angle_bins_in) where n_angle_bins in = len(angle_vector)/2

    :param th_in: Polar angle of the incoming light (in radians)
    :param phi_in: Azimuthal angle of the incoming light (in radians), or can be set as 'all' \
    in which case the power is spread equally over all the phi bins for the relevant theta.
    :param num_wl: Number of wavelengths
    :param n_theta_bins: Number of theta bins in the matrix multiplication
    :param c_azimuth: c_azimuth used to generate the matrices being multiplied
    :param phi_sym: Defines symmetry element [0, phi_sym] into which values of phi can be collapsed (in radians)

    :return: v0, an array of size (num_wl, n_angle_bins_in)
    """

    theta_intv, phi_intv, angle_vector = make_angle_vector(
        n_theta_bins, phi_sym, c_azimuth, theta_spacing
    )
    n_a_in = int(len(angle_vector) / 2)
    v0 = np.zeros((num_wl, n_a_in))
    th_bin = np.digitize(th_in, theta_intv) - 1
    phi_intv = phi_intv[th_bin]
    ov_bin = np.argmin(abs(angle_vector[:, 0] - th_bin))
    if phi_in == "all":
        n_phis = len(phi_intv) - 1
        v0[:, ov_bin : (ov_bin + n_phis)] = 1 / n_phis
    else:
        phi_ind = np.digitize(phi_in, phi_intv) - 1
        ov_bin = ov_bin + phi_ind
        v0[:, ov_bin] = 1
    return v0


def out_to_in_matrix(phi_sym, angle_vector, theta_intv, phi_intv):

    if phi_sym == 2 * np.pi:
        phi_sym = phi_sym - 0.0001
    out_to_in = np.zeros((len(angle_vector), len(angle_vector)))
    binned_theta_out = (
        np.digitize(np.pi - angle_vector[:, 1], theta_intv, right=True) - 1
    )

    phi_rebin = fold_phi(angle_vector[:, 2] + np.pi, phi_sym)

    phi_out = xr.DataArray(
        phi_rebin,
        coords={"theta_bin": (["angle_in"], binned_theta_out)},
        dims=["angle_in"],
    )

    bin_out = (
        phi_out.groupby("theta_bin")
        .map(overall_bin, args=(phi_intv, angle_vector[:, 0]))
        .data
    )

    out_to_in[bin_out, np.arange(len(angle_vector))] = 1

    up_to_down = out_to_in[int(len(angle_vector) / 2) :, : int(len(angle_vector) / 2)]
    down_to_up = out_to_in[: int(len(angle_vector) / 2), int(len(angle_vector) / 2) :]

    return COO.from_numpy(up_to_down), COO.from_numpy(down_to_up)


def make_D(alphas, thick, thetas):
    """
    Makes the bulk absorption vector for the bulk material.

    :param alphas: absorption coefficient (m^{-1})
    :param thick: thickness of the slab in m
    :param thetas: incident thetas in angle_vector (second column)

    :return:
    """
    diag = np.exp(-alphas[:, None] * thick / abs(np.cos(thetas[None, :])))
    D_1 = stack([COO.from_numpy(np.diag(x)) for x in diag])
    return D_1

# dot_wl and dot_wl_u2d now use einsum method from sparse, which is significantly
# faster than the previous for loop implementation. Thanks to Johnson Wong
# (GitHub: arsonwong)
def dot_wl(mat, vec):
    # note: sometimes get nans here in result, even when there are no nans
    # in mat or vec.
    if len(mat.shape) == 3:
        result = einsum('ijk,ik->ij', mat, COO(vec)).todense()

    if len(mat.shape) == 2:
        result = einsum('jk,ik->ij', mat, COO(vec)).todense()

    return result


def dot_wl_u2d(mat, vec):

    result = einsum('jk,ik->ij', mat, COO(vec)).todense()
    return result


def bulk_profile_calc(v_1, v_2, alphas, thetas, d, depths, A):

    per_bin = v_1 - v_2  # total absorption per bin
    abscos = np.abs(np.cos(thetas))
    denom = 1 - np.exp(-alphas[:, None] * d / abscos[None, :])
    norm = np.divide(per_bin, denom, where=denom != 0)

    result = np.empty((v_1.shape[0], len(depths)))

    for i1 in range(v_1.shape[0]):
        a_x = ((alphas[i1] * norm[i1]) / (abscos))[None, :] * np.exp(
            -alphas[i1] * depths[:, None] / abscos[None, :]
        )
        result[i1, :] = np.sum(a_x, 1)

    check = np.trapz(result, depths, axis=1)
    # the bulk layer is often thick so you don't want the depth spacing too fine,
    # but at short wavelengths this causes an issue where the integrated depth profile
    # is not equal to the total absorption. Scale the profile to fix this and make things
    # consistent.
    scale = np.divide(A, check, where=check != 0)

    corrected = scale[:, None] * result

    return corrected


def load_redistribution_matrices(
    results_path, n_a_in, n_interfaces, layer_names, front_or_rear, calc_prof_list=None
):

    R = []
    T = []
    A = []
    P = []
    I = []

    if front_or_rear == "front":
        n_max = n_interfaces

    else:
        n_max = n_interfaces - 1

    for i1 in range(n_max):

        mat_path = os.path.join(
            results_path, layer_names[i1] + front_or_rear + "RT.npz"
        )
        absmat_path = os.path.join(
            results_path, layer_names[i1] + front_or_rear + "A.npz"
        )

        fullmat = load_npz(mat_path)
        absmat = load_npz(absmat_path)

        if front_or_rear == "front":  # matrices for front incidence

            if len(fullmat.shape) == 3:
                R.append(fullmat[:, :n_a_in, :])
                T.append(fullmat[:, n_a_in:, :])
                A.append(absmat)

            else:

                R.append(fullmat[:n_a_in, :])
                T.append(fullmat[n_a_in:, :])
                A.append(absmat)

        else:  # matrices for rear incidence
            if len(fullmat.shape) == 3:
                R.append(fullmat[:, n_a_in:, :])
                T.append(fullmat[:, :n_a_in, :])
                A.append(absmat)

            else:
                R.append(fullmat[n_a_in:, :])
                T.append(fullmat[:n_a_in, :])
                A.append(absmat)

        if calc_prof_list[i1] is not None:
            profmat_path = os.path.join(
                results_path, layer_names[i1] + front_or_rear + "profmat.nc"
            )
            prof_int = xr.load_dataset(profmat_path)
            profile = prof_int["profile"]
            intgr = prof_int["intgr"]
            P.append(profile)
            I.append(intgr)

        else:
            P.append([])
            I.append([])

    return R, T, A, P, I


def append_per_pass_info(i1, vr, vt, a, vf_2, vb_1, Tb, Tf, Af, Ab):

    vr[i1].append(
        dot_wl(Tb[i1], vf_2[i1])
    )  # matrix travelling up in medium 0, i.e. reflected overall by being transmitted through front surface
    vt[i1].append(
        dot_wl(Tf[i1 + 1], vb_1[i1])
    )  # transmitted into medium below through back surface

    a[i1 + 1].append(dot_wl(Af[i1 + 1], vb_1[i1]))  # absorbed in 2nd surface
    a[i1].append(dot_wl(Ab[i1], vf_2[i1]))  # absorbed in 1st surface (from the back)

    return vr, vt, a


def matrix_multiplication(
    bulk_mats, bulk_thick, options, layer_names, calc_prof_list, save_location
):
    """

    :param bulk_mats: list of bulk materials
    :param bulk_thick: list of bulk thicknesses (in m)
    :param options: user options (State object)
    :param layer_names: list of names of the Interface layers, to load the redistribution matrices
    :param calc_prof_list: list of lists - for each interface, which layers should be included in profile calculations
           (can be empty)
    :param save_location: string, location of saved matrices
    :return:
    :rtype:
    """

    results_path = get_savepath(save_location, options.project_name)

    n_bulks = len(bulk_mats)
    n_interfaces = n_bulks + 1

    theta_spacing = (
        options.theta_spacing if "theta_spacing" in options else "sin"
    )

    theta_intv, phi_intv, angle_vector = make_angle_vector(
        options.n_theta_bins,
        options.phi_symmetry,
        options.c_azimuth,
        theta_spacing,
    )
    n_a_in = int(len(angle_vector) / 2)

    num_wl = len(options["wavelength"])

    # bulk thickness in m

    thetas = angle_vector[:n_a_in, 1]

    if options.phi_in != "all" and options.phi_in > options.phi_symmetry:
        # fold phi_in back into phi_symmetry
        phi_in = fold_phi(options["phi_in"], options["phi_symmetry"])

    else:
        phi_in = options["phi_in"]

    v0 = make_v0(
        options["theta_in"],
        phi_in,
        num_wl,
        options["n_theta_bins"],
        options["c_azimuth"],
        options["phi_symmetry"],
        theta_spacing,
    )

    up2down, down2up = out_to_in_matrix(
        options["phi_symmetry"], angle_vector, theta_intv, phi_intv
    )

    D = []
    depths_bulk = []
    for i1 in range(n_bulks):
        D.append(
            make_D(bulk_mats[i1].alpha(options["wavelength"]), bulk_thick[i1], thetas)
        )

        if options["bulk_profile"]:
            depths_bulk.append(
                np.arange(0, bulk_thick[i1], options["depth_spacing_bulk"])
            )

    # load redistribution matrices

    Rf, Tf, Af, Pf, If = load_redistribution_matrices(
        results_path, n_a_in, n_interfaces, layer_names, "front", calc_prof_list
    )

    Rb, Tb, Ab, Pb, Ib = load_redistribution_matrices(
        results_path, n_a_in, n_interfaces, layer_names, "rear", calc_prof_list
    )

    len_calcs = np.array([len(x) if x is not None else 0 for x in calc_prof_list])

    a = [[] for _ in range(n_interfaces)]
    vr = [[] for _ in range(n_bulks)]
    vt = [[] for _ in range(n_bulks)]
    A = [[] for _ in range(n_bulks)]

    vf_1 = [[] for _ in range(n_interfaces)]
    vb_1 = [[] for _ in range(n_interfaces)]
    vf_2 = [[] for _ in range(n_interfaces)]
    vb_2 = [[] for _ in range(n_interfaces)]

    if np.any(len_calcs > 0) or options.bulk_profile:
        # need to calculate profiles in either the bulk or the interfaces

        a_prof = [[] for _ in range(n_interfaces)]
        A_prof = [[] for _ in range(n_bulks)]
        logger.debug(f"Initial intensity: {np.sum(v0, axis=1)}")

        for i1 in range(n_bulks):

            # v0 is actually travelling down, but no reason to start in 'outgoing' ray format.
            vf_1[i1] = dot_wl(Tf[i1], v0)  # pass through front surface
            # print(vf_1[i1])
            # vf_1: incoming to outgoing
            # print("Transmitted through front", np.sum(vf_1[i1], axis=1))

            vr[i1].append(dot_wl(Rf[i1], v0))  # reflected from front surface
            # print("Reflected from front", np.sum(vr[i1][-1], axis=1))
            a[i1].append(
                dot_wl(Af[i1], v0)
            )  # absorbed in front surface at first interaction
            # print("Absorbed in front", np.sum(a[i1][-1], axis=1))

            if len(If[i1]) > 0:
                scale = ((np.sum(Af[i1].todense(), 1) * v0) / If[i1]).fillna(0)
                # print(((np.sum(Af[i1].todense(), 1) * v0) / If[i1]))
                # print("Af", Af[i1].todense())
                # print("v0", v0)
                # print("last few points of Pf:", Pf[i1][:, -5:, :])
                scaled_prof = scale * Pf[i1]
                a_prof[i1].append(np.sum(scaled_prof, 1))
                # print("SHAPE:", a_prof[i1][-1].shape)
                # print("Integrated absorbed:", np.trapz(a_prof[i1][-1], dx=options.depth_spacing*1e9, axis=1))

            power = np.sum(vf_1[i1], axis=1)
            # print("Power remaining", power)

            # rep
            i2 = 1

            while np.any(power > options["I_thresh"]):
                vf_1[i1] = dot_wl_u2d(down2up, vf_1[i1])  # outgoing to incoming
                # print("Travelling down int, before", np.sum(vf_1[i1], axis=1))
                # vb_1: incoming (just absorption through bulk)
                vb_1[i1] = dot_wl(D[i1], vf_1[i1])  # pass through bulk, downwards
                # vb_1 already an incoming ray
                # print("Travelling down int, after ", np.sum(vb_1[i1], axis=1))

                if len(If[i1 + 1]) > 0:

                    scale = (
                        (np.sum(Af[i1 + 1].todense(), 1) * vb_1[i1]) / If[i1 + 1]
                    ).fillna(0)
                    scaled_prof = scale * Pf[i1 + 1]
                    # print("Pf, back surface", Pf[i1+1])
                    a_prof[i1 + 1].append(np.sum(scaled_prof, 1))
                    # print("Integrated absorbed (back):", np.trapz(a_prof[i1 + 1][-1], dx=options.depth_spacing * 1e9, axis=1))

                A[i1].append(np.sum(vf_1[i1], 1) - np.sum(vb_1[i1], 1))
                # print("Total absorbed, back", A[i1][-1])

                if options.bulk_profile:
                    A_prof[i1].append(
                        bulk_profile_calc(
                            vf_1[i1],
                            vb_1[i1],
                            bulk_mats[i1].alpha(options["wavelength"]),
                            thetas,
                            bulk_thick[i1],
                            depths_bulk[i1],
                            A[i1][-1],
                        )
                    )
                    # print("bulk profile (down) integrated", np.trapz(A_prof[i1][-1], dx=options.depth_spacing_bulk, axis=1))

                vb_2[i1] = dot_wl(
                    Rf[i1 + 1], vb_1[i1]
                )  # reflect from back surface. incoming -> up
                # print("Reflected from back", np.sum(vb_2[i1], axis=1))
                # vb_2: outgoing
                vf_2[i1] = dot_wl(D[i1], vb_2[i1])  # pass through bulk, upwards
                # print("Travelling up int, after", np.sum(vf_2[i1], axis=1))

                A[i1].append(
                    np.sum(vb_2[i1], 1) - np.sum(vf_2[i1], 1)
                )  # binning doesn't matter here because summing
                # print("Bulk A (total):", A[i1][-1])

                if options.bulk_profile:
                    # vb_2 needs to be transformed to incoming ray format
                    A_prof[i1].append(
                        np.flip(
                            bulk_profile_calc(
                                vb_2[i1],
                                vf_2[i1],
                                bulk_mats[i1].alpha(options["wavelength"]),
                                thetas,
                                bulk_thick[i1],
                                depths_bulk[i1],
                                A[i1][-1],
                            ),
                            1,
                        )
                    )
                    # print("bulk profile (up) integrated", np.trapz(A_prof[i1][-1], dx=options.depth_spacing_bulk, axis=1))

                vf_2[i1] = dot_wl_u2d(up2down, vf_2[i1])  # prepare for rear incidence
                # vf_2: incoming
                # print("Travelling up, before R", np.sum(vf_2[i1], axis=1))

                if len(Ib[i1]) > 0:
                    scale = ((np.sum(Ab[i1].todense(), 1) * vf_2[i1]) / Ib[i1]).fillna(
                        0
                    )
                    scaled_prof = scale * Pb[i1]
                    # print("Pb", Pb[i1])
                    a_prof[i1].append(np.sum(scaled_prof, 1))
                    # print("SHAPE:", a_prof[i1][-1].shape)
                    # print("Integrated absorbed (front surface, rear inc):",
                    #       np.trapz(a_prof[i1][-1], dx=options.depth_spacing * 1e9, axis=1))

                vf_1[i1] = dot_wl(Rb[i1], vf_2[i1])  # reflect from front surface
                # print("Reflected from front", np.sum(vf_1[i1], axis=1))
                # vf_1 will be outgoing, gets fixed at start of next loop
                power = np.sum(vf_1[i1], axis=1)

                vr, vt, a = append_per_pass_info(
                    i1, vr, vt, a, vf_2, vb_1, Tb, Tf, Af, Ab
                )

                # print("Absorbed from back", np.sum(a[i1][-1], axis=1))

                # rewrite as f string:

                logger.info(f"After iteration {i2}: maximum power fraction remaining = {np.max(power)}")

                i2 += 1

    else:  # no profile calculation in bulk or interfaces

        for i1 in range(n_bulks):

            vf_1[i1] = dot_wl(Tf[i1], v0)  # pass through front surface
            vr[i1].append(dot_wl(Rf[i1], v0))  # reflected from front surface
            a[i1].append(
                dot_wl(Af[i1], v0)
            )  # absorbed in front surface at first interaction
            power = np.sum(vf_1[i1], axis=1)

            # vf_1[i1] = dot_wl_u2d(up2down, vf_1[i1])

            # rep
            i2 = 1

            while np.any(power > options["I_thresh"]):
                vf_1[i1] = dot_wl_u2d(down2up, vf_1[i1])  # outgoing to incoming
                vb_1[i1] = dot_wl(D[i1], vf_1[i1])  # pass through bulk, downwards
                A[i1].append(np.sum(vf_1[i1], 1) - np.sum(vb_1[i1], 1))

                vb_2[i1] = dot_wl(Rf[i1 + 1], vb_1[i1])  # reflect from back surface
                vf_2[i1] = dot_wl(D[i1], vb_2[i1])  # pass through bulk, upwards
                vf_2[i1] = dot_wl_u2d(up2down, vf_2[i1])  # prepare for rear incidence
                vf_1[i1] = dot_wl(Rb[i1], vf_2[i1])  # reflect from front surface
                A[i1].append(np.sum(vb_2[i1], 1) - np.sum(vf_2[i1], 1))
                power = np.sum(vf_1[i1], axis=1)
                logger.info(f"After iteration {i2}: maximum power fraction remaining = {np.max(power)}")

                vr, vt, a = append_per_pass_info(
                    i1, vr, vt, a, vf_2, vb_1, Tb, Tf, Af, Ab
                )

                i2 += 1

    vr = [np.array(item) for item in vr]
    vt = [np.array(item) for item in vt]
    a = [np.array(item) for item in a]
    A = [np.array(item) for item in A]

    sum_dims = ["bulk_index", "wl"]
    sum_coords = {"bulk_index": np.arange(0, n_bulks), "wl": options["wavelength"]}

    R = xr.DataArray(
        np.array([np.sum(item, (0, 2)) for item in vr]),
        dims=sum_dims,
        coords=sum_coords,
        name="R",
    )

    if i2 > 1:

        A_bulk = xr.DataArray(
            np.array([np.sum(item, 0) for item in A]),
            dims=sum_dims,
            coords=sum_coords,
            name="A_bulk",
        )

        T = xr.DataArray(
            np.array([np.sum(item, (0, 2)) for item in vt]),
            dims=sum_dims,
            coords=sum_coords,
            name="T",
        )

        if np.any(len_calcs > 0) or options.bulk_profile:

            A_prof = [np.array(item) for item in A_prof]

            a_prof = [np.array(item) for item in a_prof]

            results_per_pass = {
                "r": vr,
                "t": vt,
                "a": a,
                "A": A,
                "a_prof": a_prof,
                "A_prof": A_prof,
            }

            A_interface = xr.DataArray(
                np.array([np.sum(item, (0, 2)) for item in a]),
                dims=["surf_index", "wl"],
                coords={
                    "surf_index": np.arange(0, n_interfaces),
                    "wl": options["wavelength"],
                },
                name="A_interface",
            )
            profile = []
            for j1, item in enumerate(a_prof):
                if len(item) > 0:
                    item[item < 0] = 0
                    profile.append(
                        xr.DataArray(
                            np.sum(item, 0),
                            dims=["wl", "z"],
                            coords={"wl": options["wavelength"]},
                            name="A_profile" + str(j1),
                        )
                    )  # not necessarily same number of z coords per layer stack

            bulk_profile = [np.sum(prof_el, 0) for prof_el in A_prof]
            RAT = xr.merge([R, A_bulk, A_interface, T])

            return RAT, results_per_pass, profile, bulk_profile

        else:

            results_per_pass = {"r": vr, "t": vt, "a": a, "A": A}

            RAT = xr.merge([R, A_bulk, T])

            return RAT, results_per_pass

    else:
        RAT = xr.merge([R])
        results_per_pass = {"r": vr, "t": vt, "a": a, "A": A}

        return RAT, results_per_pass



# Copyright (C) 2021-2024 Phoebe Pearce
#
# This file is part of RayFlare and is released under the GNU Lesser General Public License (LGPL), version 3.
# Please see the LICENSE.txt file included as part of this package.
#
# Contact: p.pearce@unsw.edu.au

import numpy as np
from cmath import sqrt, acos, atan
from math import atan2
from random import random
from copy import deepcopy
from numba import jit
from scipy.spatial import Delaunay
from rayflare.utilities import process_pol

unit_cell_N = np.array(
    [[0, -1, 0], [-1, 0, 0], [0, 1, 0], [1, 0, 0]]
)  # surface normals: top, right, bottom, left

class Ray:
    """Class to store ray information for ray-tracing calculations."""
    def __init__(self,
                 intensity,
                 direction,
                 current_location,
                 s_vector,
                 p_vector,
                 pol,
                 ):
        """
        :param intensity: intensity of the ray (initial = 1)
        :param direction: tuple of the form (theta, phi) in radians
        :param current_location: 3D coordinate of the form (x, y, z)
        :param polarization: length 2 np.array of the form (s, p) where s and p are the fractions of
          the intensity in which are s and p polarized (always need to sum to 1)
        """

        self.d = direction
        self.r_a = current_location
        self.I = intensity

        self.s_vector = s_vector
        self.p_vector = p_vector

        self.pol = pol


class RTSurface:
    """Class which is used to store information about the surface which is used for ray-tracing."""

    def __init__(self, Points, interface_layers=None,
                 phong=False,
                 analytical=False,
                 **kwargs):
        """Initializes the surface.
        Parameters:

        :param Points: A numpy array of shape (n, 3) where n is the number of points on the surface. The columns are the
                        x, y and z coordinates of the points.
        :param interface_layers: a list of layers (typically, Solcore Layer objects) which are on the interface. Optional.
        :param coverage_height: The height at which the surface is expected to cover the whole unit cell. If this is not
                                provided (None), this function will try to guess the coverage height by finding the height
                                at which both the x and y coordinate are minimized.
        """

        if "height_distribution" in kwargs:
        #     return from a probability distribution instead of fixed values. This
        #     will change the height of the pyramids (assume only pyramids for now)
        #     but not the size of the unit cell. Simplices also do not change.
        #     Changing: Points, P_0s, P_1s, P_2s, crossP, N, z_min, z_max, z_cov if
        #     it's not 0 (but for pyramids it is).
            self.distribution = kwargs["height_distribution"]

        else:
            self.distribution = None

        tri = Delaunay(Points[:, [0, 1]])
        self.simplices = tri.simplices
        self.Points = Points
        self.original_Points = deepcopy(Points)
        self.height_ind = np.argmax(np.abs(self.Points[:,2]))
        self.P_0s = Points[tri.simplices[:, 0]]
        self.P_1s = Points[tri.simplices[:, 1]]
        self.P_2s = Points[tri.simplices[:, 2]]
        self.crossP = np.cross(self.P_1s - self.P_0s, self.P_2s - self.P_0s)
        self.N = self.crossP / np.linalg.norm(self.crossP, axis=1)[:, None]
        self.size = self.P_0s.shape[0]
        self.Lx = abs(min(Points[:, 0]) - max(Points[:, 0]))
        self.Ly = abs(min(Points[:, 1]) - max(Points[:, 1]))
        self.x_min = min(Points[:, 0])
        self.x_max = max(Points[:, 0])
        self.y_min = min(Points[:, 1])
        self.y_max = max(Points[:, 1])
        self.z_min = min(Points[:, 2])
        self.z_max = max(Points[:, 2])

        self.phong = phong

        if "phong_options" in kwargs:
            self.phong_options = kwargs["phong_options"]

        else:
            self.phong_options = [25, True]

        self.analytical = analytical

        if "n_analytical_interactions" in kwargs:
            self.n_analytical_interactions = kwargs["n_analytical_interactions"]

        else:
            self.n_analytical_interactions = 3

        # zcov is the height at which the surface covers the whole unit cell; i.e. it is safe to aim a ray at the unit
        # cell at this height and be sure that it will hit the surface. The method below works well for regular textures
        # like regular pyramids but doesn't work well for e.g. AFM scans, hyperhemisphere

        if "coverage_height" in kwargs:
            self.zcov = kwargs["coverage_height"]

        # catch exception here in case the surface is not regular
        else:
            self.zcov = Points[:, 2][
                np.all(
                    np.array(
                        [
                            Points[:, 0] == min(Points[:, 0]),
                            Points[:, 1] == min(Points[:, 1]),
                        ]
                    ),
                    axis=0,
                )
            ][0]

        if "name" in kwargs:
            self.name = kwargs["name"]

        else:
            self.name = ""

        if interface_layers is not None:
            self.interface_layers = interface_layers

            if "coherency_list" in kwargs:
                self.coherency_list = kwargs["coherency_list"]

            if "prof_layers" in kwargs:
                self.prof_layers = kwargs["prof_layers"]

    def __deepcopy__(self, memo):
        copy = type(self)(Points=self.Points, coverage_height=self.zcov)
        memo[id(self)] = copy

        keys = self.__dict__.keys()

        for key in keys:
            if key != "interface_layers":
                setattr(copy, key, deepcopy(getattr(self, key), memo))

        return copy

    def find_area(self):
        xyz = np.stack((self.P_0s, self.P_1s, self.P_2s))
        cos_theta = np.sum((xyz[0] - xyz[1]) * (xyz[2] - xyz[1]), 1)

        theta = np.arccos(cos_theta)
        self.area = np.sum(
            (
                0.5
                * np.linalg.norm(xyz[0] - xyz[1], axis=1)
                * np.linalg.norm(xyz[2] - xyz[1], axis=1)
                * np.sin(theta)
            )
        ) / (self.Lx * self.Ly)

    def shift(self, z_shift):
        self.Points[:, 2] = self.Points[:, 2] - z_shift
        self.P_0s = self.Points[self.simplices[:, 0]]
        self.P_1s = self.Points[self.simplices[:, 1]]
        self.P_2s = self.Points[self.simplices[:, 2]]
        self.crossP = np.cross(self.P_1s - self.P_0s, self.P_2s - self.P_0s)
        self.z_min = min(self.Points[:, 2])
        self.z_max = max(self.Points[:, 2])

        self.zcov = self.zcov - z_shift

    def refresh(self):

        if self.distribution is not None:

            new_height = np.random.choice(self.distribution["h"],
                                             p=self.distribution["p"])

            scaling = np.abs(new_height/(self.original_Points[:,2][self.height_ind] - self.zcov))

            self.Points[:,2] = scaling*(self.original_Points[:,2] - self.zcov) + self.zcov

            self.P_0s = self.Points[self.simplices[:, 0]]
            self.P_1s = self.Points[self.simplices[:, 1]]
            self.P_2s = self.Points[self.simplices[:, 2]]
            self.crossP = np.cross(self.P_1s - self.P_0s, self.P_2s - self.P_0s)
            self.N = self.crossP / np.linalg.norm(self.crossP, axis=1)[:, None]
            self.z_min = min(self.Points[:, 2])
            self.z_max = max(self.Points[:, 2])

def make_pol_vectors(pol_string, theta, phi):
    pol = process_pol(pol_string)
    pol = np.array(pol) / np.sum(pol)

    initial_p_dir = np.array([np.cos(theta) * np.cos(phi),
                              np.cos(theta) * np.sin(phi),
                              -np.sin(theta)])

    initial_s_dir = np.array([-np.sin(phi), np.cos(phi), np.zeros_like(theta)])

    initial_pol_vectors = np.array([initial_s_dir, initial_p_dir])

    return pol, initial_pol_vectors

@jit(nopython=True, error_model="numpy")
def calculate_tuv(d, tri_size, tri_crossP, r_a, tri_P_0s, tri_P_2s, tri_P_1s):

    D = (-d).repeat(tri_size).reshape((-1, tri_size)).T # this is significantly faster than np.tile
    pref = 1 / np.sum(D * tri_crossP, axis=1)
    corner = r_a - tri_P_0s

    t = pref * np.sum(tri_crossP * corner, axis=1)
    u = pref * np.sum(np.cross(tri_P_2s - tri_P_0s, D) * corner, axis=1)
    v = pref * np.sum(np.cross(D, tri_P_1s - tri_P_0s) * corner, axis=1)

    return t, u, v


@jit(nopython=True, error_model="numpy")
def normalize(x):
    return x / norm(x)

@jit(nopython=True)
def norm(x):
    s = x[0]**2 + x[1]**2 + x[2]**2
    return np.sqrt(s)

@jit(nopython=True, error_model="numpy")
def calc_intersection_properties(t, which_intersect, r_a, d, tri_N):
    t = t[which_intersect]
    ind = np.argmin(t)
    t = min(t)

    intersn = r_a + t * d

    N = tri_N[which_intersect][ind]

    Nxd = np.cross(N, -d)

    theta = atan(norm(Nxd) / np.dot(N, -d))

    return intersn, theta, N

def check_intersect(r_a, d, tri_size, tri_crossP, tri_P_0s, tri_P_2s, tri_P_1s, tri_N):

    t, u, v = calculate_tuv(d, tri_size, tri_crossP, r_a, tri_P_0s, tri_P_2s, tri_P_1s)

    # for some reason numba doesn't like this bit:
    which_intersect = (
        (u + v <= 1) & (np.all(np.vstack((u, v)) >= -1e-10, axis=0)) & (t > 0)
    )
    # get errors if set lower limit exactly to zero, hence 1e-10

    if sum(which_intersect) > 0:
        return calc_intersection_properties(t, which_intersect, r_a, d, tri_N)

    else:
        return False

def update_ray_d_pol(ray, rnd, R, T, Rs, Rp, Ts, Tp, A_per_layer, n0, n1, N, side,
                     ray_plane_s_direction, s_comp_sq):

    # TODO: is it necessary to normalize here or will it already be normalized?
    # comment out to make faster if not necessary
    if np.abs(norm(ray.d) - 1) > 1e-2:
        raise ValueError(f"Ray direction not normalized {norm(ray.d)}")

    if rnd <= R:  # REFLECTION
        ray.d = np.real(ray.d - 2 * np.dot(ray.d, N) * N)

        ray.s_vector = ray_plane_s_direction
        ray.p_vector = normalize(np.cross(ray.d, ray.s_vector))

        # the relative s/p weighting of the reflected ray needs to be weighted by
        # Rs and Rp:
        ray.pol = np.array([Rs / R, Rp / R])

        A = None

    elif (rnd > R) & (rnd <= (R + T)):  # TRANSMISSION
        # transmission, refraction
        # tr_par = (np.real(n0) / np.real(n1)) * (d - np.dot(d, N) * N)
        tr_par = (n0 / n1) * (ray.d - np.dot(ray.d, N) * N)
        tr_perp = -sqrt(1 - norm(tr_par) ** 2) * N

        side = -side
        ray.d = np.real(tr_par + tr_perp)
        ray.s_vector = ray_plane_s_direction

        ray.p_vector = normalize(np.cross(ray.d, ray.s_vector))

        # the relative s/p weighting of the transmitted ray needs to be weighted by
        # Rs and Rp:
        ray.pol = np.array([Ts / T, Tp / T])

        A = None

    else:
        # absorption
        ray.pol = np.array([s_comp_sq, 1 - s_comp_sq])
        A = A_per_layer

    ray.d = normalize(ray.d)

    return side, A

def decide_RT_Fresnel(ray, n0, n1, theta, N, side, rnd,
                        lookuptable=None, d_theta=None):

    s_component_sq, p_component_sq, ray_plane_s_direction = (
        get_pol_component_direction(theta, ray.d, N, ray.s_vector, ray.p_vector, ray.pol))

    ratio = np.clip(np.real(n1) / np.real(n0), -1, 1)

    if abs(theta) > np.arcsin(ratio):
        Rs, Rp = 1, 1

    else:
        Rs, Rp = calc_R(n0, n1, theta, np.array([s_component_sq, p_component_sq]))

    Ts = 1 - Rs
    Tp = 1 - Rp

    Rs = Rs * s_component_sq
    Rp = Rp * p_component_sq

    Ts = Ts * s_component_sq
    Tp = Tp * p_component_sq

    # R = ray.pol[0]*Rs + ray.pol[1]*Rp
    R = Rs + Rp
    T = Ts + Tp
    # R_plus_T = 1

    side, _ = update_ray_d_pol(ray, rnd, R, T, Rs, Rp, Ts, Tp, 0,
                              n0, n1, N, side, ray_plane_s_direction, s_component_sq)

    return side, None  # never absorbed, A = False


def decide_RT_TMM(ray, n0, n1, theta, N, side, rnd, lookuptable, d_theta):

    s_component_sq, p_component_sq, ray_plane_s_direction = (
        get_pol_component_direction(theta, ray.d, N, ray.s_vector, ray.p_vector, ray.pol))
    # print('s/p', s_component_sq, 1-s_component_sq)

    data_s = lookuptable.loc[dict(side=side)]

    Rs, Rp, Ts, Tp, A_per_layer = get_RT_data(theta, d_theta,
                                      data_s.R.data, data_s.T.data, data_s.Alayer.data,
                                           np.array([s_component_sq, p_component_sq]))

    R = Rs + Rp # overall probability this ray will reflect. Rs and Rp are already scaled (in
    # get_RT_data) by the incoming s and p component

    T = Ts + Tp # overall probability this ray will transmit

    side, A = update_ray_d_pol(ray, rnd, R, T, Rs, Rp, Ts, Tp,
                           A_per_layer,
                              n0, n1, N, side, ray_plane_s_direction,
                               s_component_sq)

    return side, A

#@jit(nopython=True)
def get_RT_data(theta, d_theta, R_data, T_data, Alayer_data, pol):
    # theta HAS to be positive here!
    angle_ind = int(np.floor(theta/d_theta)) # floor to avoid issues when theta = np.pi/2

    [Rs, Rp] = R_data[:, angle_ind]*pol
    [Ts, Tp] = T_data[:, angle_ind]*pol
    A_per_layer = np.sum(Alayer_data[:, angle_ind].T * pol, 1)

    return Rs, Rp, Ts, Tp, A_per_layer

@jit(nopython=True)
def get_pol_component_direction(theta, d, N, current_s_vector, current_p_vector, ray_pol):
    if theta > 1e-4:  # non-normal incidence only, otherwise cannot take cross product between ray.d and N
        ray_plane_s_direction = normalize(np.cross(d, N))
        s_component = np.array([np.dot(current_s_vector, ray_plane_s_direction),
                                np.dot(current_p_vector, ray_plane_s_direction)])
        s_component_sq = (ray_pol[0] * s_component[0] ** 2 + ray_pol[1] * s_component[1] ** 2)

        # by definition:
        p_component_sq = 1 - s_component_sq

    else:
        [s_component_sq, p_component_sq] = ray_pol
        ray_plane_s_direction = current_s_vector

    return s_component_sq, p_component_sq, ray_plane_s_direction

@jit(nopython=True)
def calc_R(n1, n2, theta, pol_comps):
    theta_t = np.arcsin((n1 / n2) * np.sin(theta))
    if pol_comps[0] == 1: # 100% s-polarized. Only need to calculate Rs, do not need to update ray.pol
        Rs = (
                np.abs(
                    (n1 * np.cos(theta) - n2 * np.cos(theta_t))
                    / (n1 * np.cos(theta) + n2 * np.cos(theta_t))
                )
                ** 2
        )
        return Rs, 0

    elif pol_comps[1] == 1: # 100% p-polarized. Only need to calculate Rp, do not need to update ray.pol
        Rp = (
                np.abs(
                    (n1 * np.cos(theta_t) - n2 * np.cos(theta))
                    / (n1 * np.cos(theta_t) + n2 * np.cos(theta))
                )
                ** 2
        )
        return 0, Rp

    else:
        Rs = (
                np.abs(
                    (n1 * np.cos(theta) - n2 * np.cos(theta_t))
                    / (n1 * np.cos(theta) + n2 * np.cos(theta_t))
                )
                ** 2
        )
        Rp = (
                np.abs(
                    (n1 * np.cos(theta_t) - n2 * np.cos(theta))
                    / (n1 * np.cos(theta_t) + n2 * np.cos(theta))
                )
                ** 2
        )

        # need to update ray.pol: if Rs > Rp, ray becomes more s-polarized and vice-versa

        return Rs, Rp

def single_cell_check(
    ray,
    ni,
    nj,
    tri,
    Lx,
    Ly,
    side,
    z_cov,
    d_theta,
    n_interactions=0,
    wl=None,
    Fr_or_TMM=0,
    lookuptable=None,
):
    decide = {0: decide_RT_Fresnel, 1: decide_RT_TMM}

    theta = 0  # needs to be assigned so no issue with return in case of miss
    # print('side', side)
    d0 = ray.d
    intersect = True
    n_ints_loop = 0
    # [top, right, bottom, left]
    n_misses = 0
    i1 = 0

    while intersect:
        i1 = i1 + 1
        with np.errstate(
            divide="ignore", invalid="ignore"
        ):  # there will be divide by 0/multiply by inf - this is fine but gives lots of warnings
            result = check_intersect(ray.r_a, ray.d, tri.size, tri.crossP, tri.P_0s,
                                         tri.P_2s, tri.P_1s, tri.N)

        if result is False:

            n_misses += 1

            o_t = np.real(acos(ray.d[2] / norm(ray.d)))
            o_p = np.real(atan2(ray.d[1], ray.d[0]))

            if np.sign(d0[2]) == np.sign(ray.d[2]):
                intersect = False
                final_res = 1

            else:
                intersect = False
                final_res = 0

            return (
                final_res,
                o_t,
                o_p,
                theta,
                n_interactions,
                side,
            )  # theta is LOCAL incidence angle (relative to texture)

        else:

            # there has been an intersection
            n_interactions += 1
            n_ints_loop += 1

            intersn = result[0]  # coordinate of the intersection (3D)

            theta = result[1]

            N = (
                result[2] * side
            )  # so angles get worked out correctly, relative to incident face normal

            if side == 1:
                n0 = ni
                n1 = nj

            else:
                n0 = nj
                n1 = ni

            rnd = random()

            side, A = decide[Fr_or_TMM](
                ray, n0, n1, abs(theta), N, side, rnd, lookuptable, d_theta
            )

            ray.r_a = np.real(
                intersn + ray.d / 1e9
            )  # this is to make sure the raytracer doesn't immediately just find the same intersection again

            if A is not None:
                # intersect = False
                # checked_translation = True
                final_res = 2
                o_t = A
                o_p = 0

                return (
                    final_res,
                    o_t,  # A array
                    o_p,
                    theta,  # LOCAL incidence angle
                    n_interactions,
                    side,
                )

@jit(nopython=True)
def exit_side(r_a, d, p_0):
    denom = np.sum(d * unit_cell_N, axis=1)
    denom[denom == 0] = 1e-12
    t = np.sum((p_0 - r_a) * unit_cell_N, axis=1) / denom  # r_intersect = r_a + t*d
    which_intersect = t > 0  # only want intersections of forward-travelling ray
    t[~which_intersect] = np.inf  # set others to inf to avoid finding when doing min
    which_side = np.argmin(t)  # find closest plane

    return which_side, t[which_side]

def single_interface_check(
    ray,
    ni,
    nj,
    tri,
    Lx,
    Ly,
    side,
    z_cov,
    d_theta,
    n_interactions=0,
    wl=None,
    Fr_or_TMM=0,
    lookuptable=None,
):
    decide = {0: decide_RT_Fresnel, 1: decide_RT_TMM}

    p_0 = np.array(
        [[0, Ly, 0], [Lx, 0, 0], [0, 0, 0], [0, 0, 0]]
    )  # points on each plane

    d0 = ray.d
    intersect = True
    checked_translation = False
    # [top, right, bottom, left]
    translation = np.array([[0, -Ly, 0], [-Lx, 0, 0], [0, Ly, 0], [Lx, 0, 0]])
    n_misses = 0
    i1 = 0

    tri.refresh()

    while intersect:
        i1 = i1 + 1
        with np.errstate(
            divide="ignore", invalid="ignore"
        ):  # there will be divide by 0/multiply by inf - this is fine but gives lots of warnings
            # result = check_intersect(ray.r_a, ray.d, tri)
            result = check_intersect(ray.r_a, ray.d, tri.size, tri.crossP, tri.P_0s,
                                         tri.P_2s, tri.P_1s, tri.N)
        # print('result (intersn, theta, N)', result)

        if result is False and not checked_translation:
            if i1 > 1:
                which_side, _ = exit_side(ray.r_a, ray.d, p_0)
                ray.r_a = ray.r_a + translation[which_side]
                # if random pyramid, need to change surface at this point
                tri.refresh()
                checked_translation = True

            else:
                if n_misses < 100:
                    # misses surface. Try again
                    if ray.d[2] < 0:  # coming from above
                        ray.r_a = np.array(
                            [
                                np.random.rand() * Lx,
                                np.random.rand() * Ly,
                                tri.z_max + 0.01,
                            ]
                        )
                    else:
                        ray.r_a = np.array(
                            [
                                np.random.rand() * Lx,
                                np.random.rand() * Ly,
                                tri.z_min - 0.01,
                            ]
                        )
                    n_misses += 1
                    i1 = 0

                else:
                    # ray keeps missing, probably because it's travelling (almost) exactly perpendicular to surface.
                    # assume it is reflected back into layer it came from
                    ray.d[2] = -ray.d[2]

                    o_t = np.real(acos(ray.d[2] / norm(ray.d)))
                    o_p = np.real(atan2(ray.d[1], ray.d[0]))
                    return 0, o_t, o_p, 0, n_interactions, side

        elif result is False and checked_translation:

            if (side == 1 and ray.d[2] < 0 and ray.r_a[2] > tri.z_min) or (
                side == -1 and ray.d[2] > 0 and ray.r_a[2] < tri.z_max
            ):
                # going down but above surface

                if ray.r_a[0] > Lx or ray.r_a[0] < 0:
                    ray.r_a[0] = (
                        ray.r_a[0] % Lx
                    )  # translate back into until cell before doing any additional translation
                if ray.r_a[1] > Ly or ray.r_a[1] < 0:
                    ray.r_a[1] = (
                        ray.r_a[1] % Ly
                    )  # translate back into until cell before doing any additional translation
                ex, t = exit_side(ray.r_a, ray.d, p_0)

                ray.r_a = ray.r_a + t * ray.d + translation[ex]
                # also change surface here
                tri.refresh()

                checked_translation = True

            else:

                o_t = np.real(acos(ray.d[2] / norm(ray.d)))
                o_p = np.real(atan2(ray.d[1], ray.d[0]))

                if np.sign(d0[2]) == np.sign(ray.d[2]):
                    intersect = False
                    final_res = 1

                else:
                    intersect = False
                    final_res = 0

                # # TODO: is this necessary?
                # if ray.r_a[0] > Lx or ray.r_a[0] < 0:
                #     ray.r_a[0] = (
                #         ray.r_a[0] % Lx
                #     )  # translate back into until cell before next ray
                # if ray.r_a[1] > Ly or ray.r_a[1] < 0:
                #     ray.r_a[1] = (
                #         ray.r_a[1] % Ly
                #     )  # translate back into until cell before next ray

                return (
                    final_res,
                    o_t,  # theta with respect to horizontal
                    o_p,
                    theta,  # LOCAL incidence angle
                    n_interactions,
                    side,
                )  # theta is LOCAL incidence angle (relative to texture)

        else:

            # there has been an intersection
            n_interactions += 1

            intersn = result[0]  # coordinate of the intersection (3D)

            theta = result[1]

            N = (
                result[2] * side
            )  # so angles get worked out correctly, relative to incident face normal

            if side == 1:
                n0 = ni
                n1 = nj

            else:
                n0 = nj
                n1 = ni

            rnd = random()

            side, A = decide[Fr_or_TMM](
                ray, n0, n1, abs(theta), N, side, rnd, lookuptable, d_theta,
            )

            ray.r_a = np.real(
                intersn + ray.d / 1e9
            )  # this is to make sure the raytracer doesn't immediately just find the same intersection again

            checked_translation = False  # reset, need to be able to translate the ray back into the unit cell again if necessary

            if A is not None:
                # intersect = False
                # checked_translation = True
                final_res = 2
                o_t = A
                o_p = 0

                return (
                    final_res,
                    o_t, # A array, NOT theta (only in the case of absorption)
                    o_p,
                    theta,  # LOCAL incidence angle
                    n_interactions,
                    side,
                )


# Copyright (C) 2021-2024 Phoebe Pearce
#
# This file is part of RayFlare and is released under the GNU Lesser General Public License (LGPL), version 3.
# Please see the LICENSE.txt file included as part of this package.
#
# Contact: p.pearce@unsw.edu.au

import numpy as np
import xarray as xr
from sparse import COO, save_npz, stack

from solcore.absorption_calculator import tmm_core_vec as tmm
from solcore.absorption_calculator import OptiStack
from solcore.state import State

from rayflare.angles import make_angle_vector, fold_phi
from rayflare.utilities import get_matrices_or_paths


def TMM(
    layers,
    incidence,
    transmission,
    surf_name,
    options,
    structpath,
    coherent=True,
    coherency_list=None,
    prof_layers=None,
    front_or_rear="front",
    save=True,
    overwrite=False,
):
    """
    Function which takes a layer stack and creates an angular redistribution matrix.

    :param layers: A list with one or more layers.
    :param incidence: incidence medium
    :param transmission: transmission medium
    :param surf_name: name of the surface (to save/load the matrices generated).
    :param options: a list of options
    :param structpath: file path where matrices will be stored or loaded from
    :param coherent: whether the layer stack is coherent. If None, it is assumed to be fully coherent. Boolean, default True.
    :param coherency_list: a list with the same number of entries as the layers, either 'c' for a coherent layer or
            'i' for an incoherent layer
    :param prof_layers: layers for which the absorption profile should be calculated
            (if None, do not calculate absorption profile at all)
    :param front_or_rear: a string, either 'front' or 'rear'; front incidence on the stack, from the incidence
            medium, or rear incidence on the stack, from the transmission medium.
    :param save: whether to save the matrices to file. Boolean, default True.
    :param overwrite: whether to overwrite existing saved matrices. Boolean, default False.

    :return: Number of returns depends on whether absorption profiles are being calculated; the first two items are
             always returned, the final one only if a profile is being calcualted.

                - fullmat: the R/T redistribution matrix at each wavelength, indexed as (wavelength, angle_bin_out, angle_bin_in)
                - A_mat: the absorption redistribution matrix (total absorption per layer), indexed as (wavelength, layer_out, angle_bin_in)
                - allres: xarray dataset storing the absorption profile data
    """

    def make_matrix_wl(wl):
        # binning into matrix, including phi
        RT_mat = np.zeros((len(theta_bins_in) * 2, len(theta_bins_in)))
        A_mat = np.zeros((n_layers, len(theta_bins_in)))

        for i1, cur_theta in enumerate(theta_bins_in):

            theta = theta_lookup[i1]  # angle_vector[i1, 1]

            data = allres.loc[dict(angle=theta, wl=wl)]

            R_prob = np.real(data["R"].data.item(0))
            T_prob = np.real(data["T"].data.item(0))

            Alayer_prob = np.real(data["Alayer"].data)
            phi_out = phis_out[i1]

            # reflection
            phi_int = phi_intv[cur_theta]
            phi_ind = np.digitize(phi_out, phi_int, right=True) - 1
            bin_out_r = np.argmin(abs(angle_vector[:, 0] - cur_theta)) + phi_ind

            RT_mat[bin_out_r, i1] = R_prob

            # transmission
            with np.errstate(divide="ignore", invalid="ignore"):
                theta_t = np.abs(
                    -np.arcsin((inc.n(wl) / trns.n(wl)) * np.sin(theta_lookup[i1]))
                    + quadrant
                )

            if np.isnan(theta_t) and T_prob > 1e-8:
                # bodge, but when transmitting into an absorbing medium, can't get total internal reflection even though
                # it is not possible to calculate the transmission angle through the method above.
                theta_t = np.abs(np.pi / 2 - 1e-5 - quadrant)

            # theta switches half-plane (th < 90 -> th >90
            if ~np.isnan(theta_t):

                theta_out_bin = np.digitize(theta_t, theta_intv, right=True) - 1
                phi_int = phi_intv[theta_out_bin]

                phi_ind = np.digitize(phi_out, phi_int, right=True) - 1
                bin_out_t = np.argmin(abs(angle_vector[:, 0] - theta_out_bin)) + phi_ind

                RT_mat[bin_out_t, i1] = T_prob

            # absorption
            A_mat[:, i1] = Alayer_prob

        fullmat = COO.from_numpy(RT_mat)
        A_mat = COO.from_numpy(A_mat)
        return fullmat, A_mat

    def make_prof_matrix_wl(wl):

        prof_wl = xr.DataArray(
            np.empty((len(dist), len(theta_bins_in))),
            dims=["z", "global_index"],
            coords={"z": dist, "global_index": np.arange(0, len(theta_bins_in))},
        )

        for i1 in range(len(theta_bins_in)):

            theta = theta_lookup[i1]

            data = allres.loc[dict(angle=theta, wl=wl)]

            prof_depth = np.real(data["Aprof"].data[0])

            prof_wl[:, i1] = prof_depth

        return prof_wl

    existing_mats, path_or_mats = get_matrices_or_paths(
        structpath, surf_name, front_or_rear, prof_layers, overwrite
    )

    if existing_mats and not overwrite:
        return path_or_mats

    else:

        wavelengths = options.wavelength

        theta_spacing = options.theta_spacing if "theta_spacing" in options else "sin"

        theta_intv, phi_intv, angle_vector = make_angle_vector(
            options.n_theta_bins,
            options.phi_symmetry,
            options.c_azimuth,
            theta_spacing,
        )
        angles_in = angle_vector[: int(len(angle_vector) / 2), :]
        thetas = np.unique(angles_in[:, 1])

        n_angles = len(thetas)

        n_layers = len(layers)

        optlayers = OptiStack(layers, substrate=transmission, incidence=incidence)
        trns = transmission
        inc = incidence

        if prof_layers is not None:
            profile = True
            z_limit = np.sum(np.array(optlayers.widths))
            full_dist = np.arange(0, z_limit, options.depth_spacing * 1e9)
            layer_start = np.insert(np.cumsum(np.insert(optlayers.widths, 0, 0)), 0, 0)
            layer_end = np.cumsum(np.insert(optlayers.widths, 0, 0))

            dist = []

            for l in prof_layers:
                dist = np.hstack(
                    (
                        dist,
                        full_dist[
                            np.all(
                                (full_dist >= layer_start[l], full_dist < layer_end[l]),
                                0,
                            )
                        ],
                    )
                )

            if front_or_rear != "front":
                dist = (z_limit - dist)[::-1]
                prof_layers = np.sort(len(layers) - np.array(prof_layers) + 1).tolist()

        else:
            profile = False
            dist = None

        if front_or_rear != "front":
            optlayers = OptiStack(
                layers[::-1], substrate=incidence, incidence=transmission
            )
            trns = incidence
            inc = transmission

        if options.pol == "u":
            pols = ["s", "p"]

        else:
            pols = [options.pol]

        R = xr.DataArray(
            np.empty((len(pols), len(wavelengths), n_angles)),
            dims=["pol", "wl", "angle"],
            coords={"pol": pols, "wl": wavelengths, "angle": thetas},
            name="R",
        )
        T = xr.DataArray(
            np.empty((len(pols), len(wavelengths), n_angles)),
            dims=["pol", "wl", "angle"],
            coords={"pol": pols, "wl": wavelengths, "angle": thetas},
            name="T",
        )

        Alayer = xr.DataArray(
            np.empty((len(pols), n_angles, len(wavelengths), n_layers)),
            dims=["pol", "angle", "wl", "layer"],
            coords={
                "pol": pols,
                "wl": wavelengths,
                "angle": thetas,
                "layer": range(1, n_layers + 1),
            },
            name="Alayer",
        )

        theta_t = xr.DataArray(
            np.empty((len(pols), len(wavelengths), n_angles)),
            dims=["pol", "wl", "angle"],
            coords={"pol": pols, "wl": wavelengths, "angle": thetas},
            name="theta_t",
        )

        if profile:
            Aprof = xr.DataArray(
                np.empty((len(pols), n_angles, len(wavelengths), len(dist))),
                dims=["pol", "angle", "wl", "z"],
                coords={"pol": pols, "wl": wavelengths, "angle": thetas, "z": dist},
                name="Aprof",
            )

        R_loop = np.empty((len(wavelengths), n_angles))
        T_loop = np.empty((len(wavelengths), n_angles))
        Alayer_loop = np.empty((n_angles, len(wavelengths), n_layers))
        th_t_loop = np.empty((len(wavelengths), n_angles))

        if profile:
            Aprof_loop = np.empty((n_angles, len(wavelengths), len(dist)))

        tmm_struct = tmm_structure(optlayers, incidence, transmission, False)

        pass_options = {}
        pass_options["coherent"] = coherent
        pass_options["coherency_list"] = coherency_list
        pass_options["wavelength"] = wavelengths
        pass_options["depth_spacing"] = options.depth_spacing

        for pol in pols:

            for i3, theta in enumerate(thetas):

                pass_options["pol"] = pol
                pass_options["theta_in"] = theta

                res = tmm_struct.calculate(
                    pass_options, profile=profile, layers=prof_layers, dist=dist
                )

                R_loop[:, i3] = np.real(res["R"])
                T_loop[:, i3] = np.real(res["T"])
                Alayer_loop[i3, :, :] = np.real(res["A_per_layer"])

                if profile:
                    Aprof_loop[i3, :, :] = res["profile"]

            # sometimes get very small negative values (like -1e-20)
            R_loop[R_loop < 0] = 0
            T_loop[T_loop < 0] = 0
            Alayer_loop[Alayer_loop < 0] = 0

            if profile:
                Aprof_loop[Aprof_loop < 0] = 0

            if front_or_rear == "rear":
                Alayer_loop = np.flip(Alayer_loop, axis=2)

                if profile:
                    Aprof_loop = np.flip(Aprof_loop, axis=2)

            R.loc[dict(pol=pol)] = R_loop
            T.loc[dict(pol=pol)] = T_loop
            Alayer.loc[dict(pol=pol)] = Alayer_loop
            theta_t.loc[dict(pol=pol)] = th_t_loop

            if profile:
                Aprof.loc[dict(pol=pol)] = Aprof_loop
                Aprof.transpose("pol", "wl", "angle", "z")

        Alayer = Alayer.transpose("pol", "wl", "angle", "layer")

        if profile:
            allres = xr.merge([R, T, Alayer, Aprof])
        else:
            allres = xr.merge([R, T, Alayer])

        if options.pol == "u":
            allres = (
                allres.reduce(np.mean, "pol").assign_coords(pol="u").expand_dims("pol")
            )

        # populate matrices
        if front_or_rear == "front":

            angle_vector_th = angle_vector[: int(len(angle_vector) / 2), 1]
            angle_vector_phi = angle_vector[: int(len(angle_vector) / 2), 2]

            phis_out = fold_phi(angle_vector_phi + np.pi, options.phi_symmetry)
            theta_lookup = angles_in[:, 1]
            quadrant = np.pi

        else:
            angle_vector_th = angle_vector[int(len(angle_vector) / 2) :, 1]
            angle_vector_phi = angle_vector[int(len(angle_vector) / 2) :, 2]

            phis_out = fold_phi(angle_vector_phi + np.pi, options.phi_symmetry)
            theta_lookup = angles_in[:, 1][::-1]
            quadrant = 0

        phis_out[phis_out == 0] = 1e-10

        theta_bins_in = np.digitize(angle_vector_th, theta_intv, right=True) - 1

        mats = [make_matrix_wl(wl) for wl in wavelengths]

        fullmat = stack([item[0] for item in mats])
        A_mat = stack([item[1] for item in mats])

        if save:
            save_npz(path_or_mats[0], fullmat)
            save_npz(path_or_mats[1], A_mat)

        if profile:
            prof_mat = [make_prof_matrix_wl(wl) for wl in wavelengths]

            profile = xr.concat(prof_mat, "wl")
            intgr = xr.DataArray(
                np.sum(A_mat.todense(), 1),
                dims=["wl", "global_index"],
                coords={
                    "wl": wavelengths,
                    "global_index": np.arange(0, len(theta_bins_in)),
                },
            )
            intgr.name = "intgr"
            profile.name = "profile"
            allres = xr.merge([intgr, profile])

            if save:
                allres.to_netcdf(path_or_mats[2])

            return fullmat, A_mat, allres

        return fullmat, A_mat


class tmm_structure:
    """Set up structure for TMM calculations.

    :param stack: an OptiStack or SolarCell object, or a list of Solcore layers.
    :param incidence: incidence medium (Solcore material)
    :param transmission: transmission medium/substrate (Solcore material)
    :param no_back_reflection: whether to suppress reflections at the interface between the final material
            in the stack and the substrate (default False)
    """

    def __init__(
        self, layer_stack, incidence=None, transmission=None, no_back_reflection=False
    ):

        if "OptiStack" in str(type(layer_stack)):
            layer_stack.no_back_reflection = no_back_reflection
        else:
            layer_stack = OptiStack(
                layer_stack,
                no_back_reflection=no_back_reflection,
                substrate=transmission,
                incidence=incidence,
            )

        self.layer_stack = layer_stack
        self.no_back_reflection = no_back_reflection
        self.width = np.sum(layer_stack.widths) / 1e9

    def calculate(self, options, profile=False, layers=None, dist=None):
        """ Calculates the reflected, absorbed and transmitted intensity of the structure for the wavelengths and angles
        defined.

        :param options: options for the calculation. The key entries are:

            - wavelength: Wavelengths (in m) in which calculate the data. An array.
            - theta_in: Angle (in radians) of the incident light.
            - pol: Polarisation of the light: 's', 'p' or 'u'.
            - coherent: If the light is coherent or not. If not, a coherency list must be added.
            - coherency_list: A list indicating in which layers light should be treated as coherent ('c') and in which \
                incoherent ('i'). It needs as many elements as layers in the structure.

        :param profile: whether or not to calculate the absorption profile
        :param layers: indices of the layers in which to calculate the absorption profile.
            Layer 0 is the incidence medium.

        :return: A dictionary with the R, A and T at the specified wavelengths and angle.
        """

        def calculate_profile(layers, dist=None):
            # layer indices: 0 is incidence, n is transmission medium

            if layers is None:
                layers = np.arange(1, layer_stack.num_layers + 1)

            if dist is None:
                depth_spacing = options.depth_spacing * 1e9  # convert from m to nm
                z_limit = np.sum(np.array(layer_stack.widths))
                full_dist = np.arange(0, z_limit, depth_spacing)
                layer_start = np.insert(
                    np.cumsum(np.insert(layer_stack.widths, 0, 0)), 0, 0
                )
                layer_end = np.cumsum(np.insert(layer_stack.widths, 0, 0))

                dist = []

                for l in layers:
                    dist = np.hstack(
                        (
                            dist,
                            full_dist[
                                np.all(
                                    (
                                        full_dist >= layer_start[l],
                                        full_dist < layer_end[l],
                                    ),
                                    0,
                                )
                            ],
                        )
                    )

            if pol in "sp":

                if coherent:
                    fn = tmm.absorp_analytic_fn().fill_in(out, layers)
                    layer, d_in_layer = tmm.find_in_structure_with_inf(
                        layer_stack.get_widths(), dist
                    )
                    data = tmm.position_resolved(layer, d_in_layer, out)
                    output["profile"] = data["absor"]

                else:
                    fraction_reaching = 1 - np.cumsum(A_per_layer, axis=0)
                    fn = tmm.absorp_analytic_fn()
                    fn.a1, fn.a3, fn.A1, fn.A2, fn.A3 = (
                        np.empty((0, num_wl)),
                        np.empty((0, num_wl)),
                        np.empty((0, num_wl)),
                        np.empty((0, num_wl)),
                        np.empty((0, num_wl)),
                    )

                    layer, d_in_layer = tmm.find_in_structure_with_inf(
                        layer_stack.get_widths(), dist
                    )
                    data = tmm.inc_position_resolved(
                        layer,
                        d_in_layer,
                        out,
                        coherency_list,
                        4
                        * np.pi
                        * np.imag(layer_stack.get_indices(wavelength))
                        / wavelength,
                    )
                    output["profile"] = data

                    for l in layers:

                        if coherency_list[l] == "c":
                            fn_l = tmm.inc_find_absorp_analytic_fn(l, out)
                            fn.a1 = np.vstack((fn.a1, fn_l.a1))
                            fn.a3 = np.vstack((fn.a3, fn_l.a3))
                            fn.A1 = np.vstack((fn.A1, fn_l.A1))
                            fn.A2 = np.vstack((fn.A2, fn_l.A2))
                            fn.A3 = np.vstack((fn.A3, fn_l.A3))

                        else:
                            alpha = (
                                np.imag(layer_stack.get_indices(wavelength)[l])
                                * 4
                                * np.pi
                                / wavelength
                            )
                            fn.a1 = np.vstack((fn.a1, alpha))
                            fn.A2 = np.vstack((fn.A2, alpha * fraction_reaching[l - 1]))
                            fn.a3 = np.vstack((fn.a3, np.zeros((1, num_wl))))
                            fn.A1 = np.vstack((fn.A1, np.zeros((1, num_wl))))
                            fn.A3 = np.vstack((fn.A3, np.zeros((1, num_wl))))

            else:
                if coherent:
                    fn_s = tmm.absorp_analytic_fn().fill_in(out_s, layers)
                    fn_p = tmm.absorp_analytic_fn().fill_in(out_p, layers)
                    fn = fn_s.add(fn_p).scale(0.5)

                    layer, d_in_layer = tmm.find_in_structure_with_inf(
                        layer_stack.get_widths(), dist
                    )
                    data_s = tmm.position_resolved(layer, d_in_layer, out_s)
                    data_p = tmm.position_resolved(layer, d_in_layer, out_p)

                    output["profile"] = 0.5 * (data_s["absor"] + data_p["absor"])

                else:
                    fraction_reaching_s = 1 - np.cumsum(A_per_layer_s, axis=0)
                    fraction_reaching_p = 1 - np.cumsum(A_per_layer_s, axis=0)
                    fraction_reaching = 0.5 * (
                        fraction_reaching_s + fraction_reaching_p
                    )
                    fn = tmm.absorp_analytic_fn()
                    fn.a1, fn.a3, fn.A1, fn.A2, fn.A3 = (
                        np.empty((0, num_wl)),
                        np.empty((0, num_wl)),
                        np.empty((0, num_wl)),
                        np.empty((0, num_wl)),
                        np.empty((0, num_wl)),
                    )

                    layer, d_in_layer = tmm.find_in_structure_with_inf(
                        layer_stack.get_widths(), dist
                    )
                    data_s = tmm.inc_position_resolved(
                        layer,
                        d_in_layer,
                        out_s,
                        coherency_list,
                        4
                        * np.pi
                        * np.imag(layer_stack.get_indices(wavelength))
                        / wavelength,
                    )
                    data_p = tmm.inc_position_resolved(
                        layer,
                        d_in_layer,
                        out_p,
                        coherency_list,
                        4
                        * np.pi
                        * np.imag(layer_stack.get_indices(wavelength))
                        / wavelength,
                    )

                    output["profile"] = 0.5 * (data_s + data_p)

                    for l in layers:
                        if coherency_list[l] == "c":
                            fn_s = tmm.inc_find_absorp_analytic_fn(l, out_s)
                            fn_p = tmm.inc_find_absorp_analytic_fn(l, out_s)
                            fn_l = fn_s.add(fn_p).scale(0.5)
                            fn.a1 = np.vstack((fn.a1, fn_l.a1))
                            fn.a3 = np.vstack((fn.a3, fn_l.a3))
                            fn.A1 = np.vstack((fn.A1, fn_l.A1))
                            fn.A2 = np.vstack((fn.A2, fn_l.A2))
                            fn.A3 = np.vstack((fn.A3, fn_l.A3))

                        else:
                            alpha = (
                                np.imag(layer_stack.get_indices(wavelength)[l])
                                * 4
                                * np.pi
                                / wavelength
                            )
                            fn.a1 = np.vstack((fn.a1, alpha))
                            fn.A2 = np.vstack((fn.A2, alpha * fraction_reaching[l - 1]))
                            fn.a3 = np.vstack((fn.a3, np.zeros((1, num_wl))))
                            fn.A1 = np.vstack((fn.A1, np.zeros((1, num_wl))))
                            fn.A3 = np.vstack((fn.A3, np.zeros((1, num_wl))))

            output["profile"][output["profile"] < 0] = 0
            output["profile_coeff"] = np.stack(
                (fn.A1, fn.A2, np.real(fn.A3), np.imag(fn.A3), fn.a1, fn.a3)
            )  # shape is (6, n_layers, num_wl)

        if isinstance(options, dict):
            options = State(options)

        wavelength = options.wavelength * 1e9
        pol = options.pol
        angle = options.theta_in

        coherent = options.coherent if "coherent" in options.keys() else True

        layer_stack = self.layer_stack

        if not coherent:
            coherency_list = self.build_coh_list(options)

        num_wl = len(wavelength)
        output = {
            "R": np.zeros(num_wl),
            "A": np.zeros(num_wl),
            "T": np.zeros(num_wl),
            "all_p": [],
            "all_s": [],
        }

        if pol in "sp":
            if coherent:
                out = tmm.coh_tmm(
                    pol,
                    layer_stack.get_indices(wavelength),
                    layer_stack.get_widths(),
                    angle,
                    wavelength,
                )
                A_per_layer = tmm.absorp_in_each_layer(out)
                output["R"] = out["R"]
                output["A"] = 1 - out["R"] - out["T"]
                output["T"] = out["T"]
                output["A_per_layer"] = A_per_layer[1:-1]
            else:
                out = tmm.inc_tmm(
                    pol,
                    layer_stack.get_indices(wavelength),
                    layer_stack.get_widths(),
                    coherency_list,
                    angle,
                    wavelength,
                )

                A_per_layer = np.array(tmm.inc_absorp_in_each_layer(out))
                output["R"] = out["R"]
                output["A"] = 1 - out["R"] - out["T"]

                # make sure everything adds to 1:
                A_per_layer[A_per_layer < 0] = 0
                output["A"][output["A"] < 0] = 0

                A_per_layer = np.divide(
                    A_per_layer[1:-1] * output["A"],
                    np.sum(A_per_layer[1:-1], axis=0),
                    where=np.sum(A_per_layer[1:-1], axis=0) != 0,
                    out=A_per_layer[1:-1],
                )

                output["T"] = out["T"]
                output["A_per_layer"] = A_per_layer
        else:
            if coherent:
                out_p = tmm.coh_tmm(
                    "p",
                    layer_stack.get_indices(wavelength),
                    layer_stack.get_widths(),
                    angle,
                    wavelength,
                )
                out_s = tmm.coh_tmm(
                    "s",
                    layer_stack.get_indices(wavelength),
                    layer_stack.get_widths(),
                    angle,
                    wavelength,
                )
                A_per_layer_p = tmm.absorp_in_each_layer(out_p)
                A_per_layer_s = tmm.absorp_in_each_layer(out_s)
                output["R"] = 0.5 * (out_p["R"] + out_s["R"])
                output["T"] = 0.5 * (out_p["T"] + out_s["T"])
                output["A"] = 1 - output["R"] - output["T"]
                output["A_per_layer"] = 0.5 * (
                    A_per_layer_p[1:-1] + A_per_layer_s[1:-1]
                )

            else:
                out_p = tmm.inc_tmm(
                    "p",
                    layer_stack.get_indices(wavelength),
                    layer_stack.get_widths(),
                    coherency_list,
                    angle,
                    wavelength,
                )
                out_s = tmm.inc_tmm(
                    "s",
                    layer_stack.get_indices(wavelength),
                    layer_stack.get_widths(),
                    coherency_list,
                    angle,
                    wavelength,
                )

                A_per_layer_p = np.array(tmm.inc_absorp_in_each_layer(out_p))
                A_per_layer_s = np.array(tmm.inc_absorp_in_each_layer(out_s))

                output["R"] = 0.5 * (out_p["R"] + out_s["R"])
                output["T"] = 0.5 * (out_p["T"] + out_s["T"])
                output["A"] = 1 - output["R"] - output["T"]
                output["all_p"] = out_p["power_entering_list"]
                output["all_s"] = out_s["power_entering_list"]
                A_per_layer = 0.5 * (A_per_layer_p[1:-1] + A_per_layer_s[1:-1])
                output["A_per_layer"] = np.divide(
                    A_per_layer * output["A"],
                    np.sum(A_per_layer, axis=0),
                    where=np.sum(A_per_layer, axis=0) != 0,
                )

        output["A_per_layer"] = output["A_per_layer"].T

        if profile:
            calculate_profile(layers, dist)

        return output

    def calculate_profile(self, options, layers=None):

        prof = self.calculate(options, profile=True, layers=layers)
        return prof

    def set_widths(self, new_widths):

        self.layer_stack.set_widths(new_widths)

    def build_coh_list(self, options):

        coherency_list = (
            options.coherency_list if "coherency_list" in options.keys() else None
        )
        if coherency_list is not None:
            assert len(coherency_list) == self.layer_stack.num_layers, (
                "Error: The coherency list (passed in the options) must have as many elements (now {}) as the "
                "number of layers (now {}).".format(
                    len(coherency_list), self.layer_stack.num_layers
                )
            )

            if self.no_back_reflection:
                coherency_list = ["i"] + coherency_list + ["i", "i"]
            else:
                coherency_list = ["i"] + coherency_list + ["i"]

            return coherency_list

        else:
            raise Exception(
                "Error: For incoherent or partly incoherent calculations you must supply the "
                "coherency_list parameter with as many elements as the number of layers in the "
                "structure"
            )



# Copyright (C) 2021-2024 Phoebe Pearce
#
# This file is part of RayFlare and is released under the GNU Lesser General Public License (LGPL), version 3.
# Please see the LICENSE.txt file included as part of this package.
#
# Contact: p.pearce@unsw.edu.au

import numpy as np
from sparse import COO, save_npz
from rayflare.angles import fold_phi
from rayflare.utilities import get_matrices_or_paths


def lambertian_matrix(
    angle_vector,
    theta_intv,
    surf_name,
    structpath,
    front_or_rear="front",
    save=True,
    overwrite=False,
):
    """
    Generates a redistribution matrix for perfect Lambertian reflection.

    :param angle_vector: an angle_vector in the standard format
    :param theta_intv: the theta (polar) intervals (edges of the bins) to be used
    :param surf_name: surface name (for saving the matrix)
    :param structpath: file path where matrices will be stored or loaded from
    :param front_or_rear: generate matrix for 'front' or 'rear' incidence?
    :param save: Boolean, whether to save the resulting matrix (True) or only return it (False). Default True
    :param overwrite: Boolean, whether to overwrite existing saved matrices. Default False

    :return: the redistribution matrix in sparse COO format.
    """

    existing_mats, path_or_mats = get_matrices_or_paths(
        structpath, surf_name, front_or_rear, overwrite=overwrite
    )

    if existing_mats and not overwrite:
        return path_or_mats

    else:
        theta_values = np.unique(angle_vector[angle_vector[:, 1] < np.pi / 2, 1])
        dtheta = np.diff(theta_intv[theta_intv <= np.pi / 2])

        dP = np.cos(theta_values) * dtheta

        # matrix has indexing (out, in): row picks out 'out' entry, column picks out which v0 element

        # since it doesn't matter what the incidence angle is for Lambertian scattering, all the columns rows be identical!

        # how many phi entries are there for each theta?

        n_phis = [np.sum(angle_vector[:, 1] == theta) for theta in theta_values]

        column = [
            x
            for sublist in [[dP[i1] / n] * n for i1, n in enumerate(n_phis)]
            for x in sublist
        ]

        whole_matrix = np.vstack([column] * int(len(angle_vector) / 2)).T

        # renormalize (rounding errors)

        whole_matrix_R = whole_matrix / np.sum(whole_matrix, 0)

        whole_matrix_T = np.zeros_like(whole_matrix_R)

        whole_matrix = np.vstack([whole_matrix_R, whole_matrix_T])

        A_matrix = np.zeros((1, int(len(angle_vector) / 2)))

        allArray = COO.from_numpy(whole_matrix)
        absArray = COO.from_numpy(A_matrix)

        if save:
            save_npz(path_or_mats[0], allArray)
            save_npz(path_or_mats[1], absArray)

        return allArray, absArray


def mirror_matrix(
    angle_vector,
    theta_intv,
    phi_intv,
    surf_name,
    options,
    structpath,
    front_or_rear="front",
    save=True,
    overwrite=False,
):
    """
    Generates a redistribution matrix for a perfect mirror (100% reflection).

    :param angle_vector: an angle_vector in the standard format
    :param theta_intv: the theta (polar) intervals (edges of the bins) to be used
    :param phi_intv: the phi (azimuthal) intervals (edges of the bins)
    :param surf_name: surface name (for saving the matrix)
    :param options: dictionary of user options
    :param structpath: file path where matrices will be stored or loaded from
    :param front_or_rear: generate matrix for 'front' or 'rear' incidence?
    :param save: Boolean, whether to save the resulting matrix (True) or only return it (False). Default True
    :param overwrite: Boolean, whether to overwrite existing saved matrices. Default False

    :return: the redistribution matrix in sparse COO format.
    """

    existing_mats, path_or_mats = get_matrices_or_paths(
        structpath, surf_name, front_or_rear, overwrite=overwrite
    )

    if existing_mats and not overwrite:
        return path_or_mats

    else:
        if front_or_rear == "front":

            angle_vector_th = angle_vector[: int(len(angle_vector) / 2), 1]
            angle_vector_phi = angle_vector[: int(len(angle_vector) / 2), 2]

            phis_out = fold_phi(angle_vector_phi + np.pi, options["phi_symmetry"])

        else:
            angle_vector_th = angle_vector[int(len(angle_vector) / 2) :, 1]
            angle_vector_phi = angle_vector[int(len(angle_vector) / 2) :, 2]

            phis_out = fold_phi(angle_vector_phi + np.pi, options["phi_symmetry"])

        # matrix will be all zeros with just one '1' in each column/row. Just need to determine where it goes

        binned_theta = np.digitize(angle_vector_th, theta_intv, right=True) - 1

        bin_in = np.arange(len(angle_vector_phi))

        phi_ind = [
            np.digitize(phi, phi_intv[binned_theta[i1]], right=True) - 1
            for i1, phi in enumerate(phis_out)
        ]
        overall_bin = [
            np.argmin(abs(angle_vector[:, 0] - binned_theta[i1])) + phi_i
            for i1, phi_i in enumerate(phi_ind)
        ]

        whole_matrix = np.zeros((len(overall_bin) * 2, len(overall_bin)))

        whole_matrix[overall_bin, bin_in] = 1

        A_matrix = np.zeros((1, len(overall_bin)))

        allArray = COO.from_numpy(whole_matrix)
        absArray = COO.from_numpy(A_matrix)

        if save:
            save_npz(path_or_mats[0], allArray)
            save_npz(path_or_mats[1], absArray)

        return allArray, absArray



# Copyright (C) 2021-2024 Phoebe Pearce
#
# This file is part of RayFlare and is released under the GNU Lesser General Public License (LGPL), version 3.
# Please see the LICENSE.txt file included as part of this package.
#
# Contact: p.pearce@unsw.edu.au

import numpy as np
import xarray as xr
import matplotlib.pyplot as plt
import seaborn as sns
import matplotlib as mpl


def make_angle_vector(n_angle_bins, phi_sym, c_azimuth, theta_spacing="sin"):
    """Makes the binning intervals & angle vector depending on the relevant options.
    :param n_angle_bins: number of bins per 90 degrees in the polar direction (theta)
    :param phi_sym: phi angle (in radians) for the rotational symmetry of the unit cell; e.g. for a square-based pyramid,
    pi/4
    :param c_azimuth: a number between 0 and 1 which determines how finely the space is discretized in the
    azimuthal direction. N_azimuth = c_azimuth*r where r is the index of the polar angle bin. N_azimuth is
    rounded up to the nearest intergest if c_azimuth is not an integer.
    :return theta_intv: edges of the theta (polar angle) bins
    :return phi_intv: list with the edges of the phi bins for every theta bin
    :return angle_vector: array where the first column is the r index (theta bin), the second column in
    the mean theta for that bin, and the third column is the mean phi for that bin.
    """

    if theta_spacing == "sin":
        sin_a_b = np.linspace(
            0, 1, n_angle_bins + 1
        )  # number of bins is between 0 and 90 degrees
        # even spacing in terms of sin(theta) rather than theta
        # will have the same number of bins between 90 and 180 degrees

        theta_intv = np.concatenate(
            [np.arcsin(sin_a_b), np.pi - np.flip(np.arcsin(sin_a_b[:-1]))]
        )

    elif theta_spacing == "linear":
        theta_intv = np.linspace(0, np.pi / 2, n_angle_bins + 1)

        theta_intv = np.concatenate([theta_intv, np.pi - np.flip(theta_intv[:-1])])

    theta_middle = (theta_intv[:-1] + theta_intv[1:]) / 2
    phi_intv = []
    angle_vector = np.empty((0, 3))

    for i1, theta in enumerate(theta_middle):
        if theta > np.pi / 2:
            ind = len(theta_intv) - (i1 + 1)  # + 1 because Python is zero-indexed
        else:
            ind = i1 + 1

        phi_intv.append(np.linspace(0, phi_sym, int(np.ceil(c_azimuth * ind) + 1)))
        phi_middle = (phi_intv[i1][:-1] + phi_intv[i1][1:]) / 2

        angle_vector = np.append(
            angle_vector,
            np.array(
                [
                    np.array(len(phi_middle) * [i1]),
                    np.array(len(phi_middle) * [theta]),
                    phi_middle,
                ]
            ).T,
            axis=0,
        )

    return theta_intv, phi_intv, angle_vector


def fold_phi(phis, phi_sym):
    """'Folds' phi angles back into symmetry element from 0 -> phi_sym radians"""
    return (abs(phis // np.pi) * 2 * np.pi + phis) % phi_sym


def theta_summary(out_mat, angle_vector, n_theta_bins, front_or_rear="front"):
    """
    Accepts an RT redistribution matrix and sums it over all the azimuthal angle bins to create an output
    in terms of theta_in and theta_out.
    :param out_mat: an RT (or just R or T) redistribution matrix
    :param angle_vector: corresponding angle_vector array (output from make_angle_vector)
    :return sum_mat: the theta summary matrix
    :return R: the overall reflection probability for every incidence theta
    :return T: the overall transmission probaility for every incidence theta
    """

    theta_all = np.unique(angle_vector[:, 1])
    # theta_r = theta_all[:n_theta_bins]
    theta_t = theta_all[n_theta_bins:]

    if front_or_rear == "front":
        out_mat = xr.DataArray(
            out_mat,
            dims=["index_out", "index_in"],
            coords={
                "theta_in": (["index_in"], angle_vector[: out_mat.shape[1], 1]),
                "theta_out": (["index_out"], angle_vector[: out_mat.shape[0], 1]),
            },
        )

    else:
        out_mat = xr.DataArray(
            out_mat,
            dims=["index_out", "index_in"],
            coords={
                "theta_in": (["index_in"], angle_vector[out_mat.shape[1] :, 1]),
                "theta_out": (["index_out"], angle_vector[: out_mat.shape[0], 1]),
            },
        )

    sum_mat = out_mat.groupby("theta_in").map(np.mean, args=(1, None))

    sum_mat = sum_mat.groupby("theta_out").map(
        weighted_mean, args=("theta_out", 0, None)
    )

    if front_or_rear == "front":
        sum_mat = xr.DataArray(
            sum_mat.data,
            dims=[r"$\theta_{out}$", r"$\theta_{in}$"],
            coords={
                r"$\theta_{out}$": theta_all,
                r"$\theta_{in}$": theta_all[: sum_mat.shape[1]],
            },
        )

    else:
        sum_mat = xr.DataArray(
            sum_mat.data,
            dims=[r"$\theta_{out}$", r"$\theta_{in}$"],
            coords={r"$\theta_{out}$": theta_all, r"$\theta_{in}$": theta_t},
        )

    return sum_mat


def weighted_mean(x, summing_over, axis, dtype=None):
    # print(x.coords[summing_over])
    # print(len(x.coords[summing_over]))
    mean = np.mean(x, axis, dtype) * len(x.coords[summing_over])
    return mean


def plot_theta_summary(summat, summat_back, n_points=100):

    whole_mat = xr.concat((summat, summat_back), dim=r"$\theta_{in}$")

    whole_mat_imshow = whole_mat.rename(
        {r"$\theta_{in}$": "theta_in", r"$\theta_{out}$": "theta_out"}
    )

    whole_mat_imshow = whole_mat_imshow.interp(
        theta_in=np.linspace(0, np.pi, n_points),
        theta_out=np.linspace(0, np.pi, n_points),
    )

    whole_mat_imshow = whole_mat_imshow.rename(
        {"theta_in": r"$\theta_{in}$", "theta_out": r"$\theta_{out}$"}
    )

    palhf = sns.cubehelix_palette(256, start=0.5, rot=-0.9)
    palhf.reverse()
    seamap = mpl.colors.ListedColormap(palhf)

    fig = plt.figure()
    ax = plt.subplot(111)
    ax = whole_mat_imshow.plot.imshow(ax=ax, cmap=seamap)
    # ax = plt.subplot(212)
    fig.savefig("matrix.png", bbox_inches="tight", format="png")
    # ax = Tth.plot.imshow(ax=ax)

    plt.show()


def theta_summary_A(A_mat, angle_vector):
    """Accepts an absorption per layer redistribution matrix and sums it over all the azimuthal angle bins to create an output
    in terms of theta_in and theta_out.
    :param out_mat: an absorption redistribution matrix
    :param angle_vector: corresponding angle_vector array (output from make_angle_vector)
    :return sum_mat: the theta summary matrix
    """
    A_mat = xr.DataArray(
        A_mat,
        dims=["layer_out", "index_in"],
        coords={
            "theta_in": (["index_in"], angle_vector[: A_mat.shape[1], 1]),
            "layer_out": 1 + np.arange(A_mat.shape[0]),
        },
    )
    sum_mat = A_mat.groupby("theta_in").map(np.sum, args=(1, None))

    return sum_mat.data


def overall_bin(x, phi_intv, angle_vector_0):
    phi_ind = np.digitize(x, phi_intv[x.coords["theta_bin"].data[0]], right=True) - 1
    ov_bin = np.argmin(abs(angle_vector_0 - x.coords["theta_bin"].data[0])) + phi_ind
    return ov_bin



import numpy as np
import xarray as xr
import os
from solcore.state import State
from .rt_common import make_pol_vectors

from rayflare import logger

theta_lamb = np.linspace(0, 0.999 * np.pi / 2, 100)
def traverse_vectorised(width, theta, alpha, I_i, positions, direction):

    ratio = alpha[None, :] / np.real(np.abs(np.cos(theta)))
    DA_u = I_i[:, :, None] * ratio[:, :, None] * np.exp((-ratio[:, :, None] * positions[None, None, :]))
    # DA_u dimensions: (directions, wavelength, position)

    I_back = I_i * np.exp(-ratio * width)

    # stop = np.where(I_back < I_thresh)[0]

    if direction == -1:
        DA_u = np.flip(DA_u)

    intgr = np.trapz(DA_u, positions, axis=2)

    DA = np.divide(
        ((I_i[:, :, None] - I_back[:, :, None]) * DA_u), intgr[:, :, None], where=intgr[:, :, None] != 0,
        out=np.zeros_like(DA_u),
    ).T

    return DA, I_back

def calc_RAT_Fresnel(theta, pol, *args):
    n1 = args[0]
    n2 = args[1]
    theta_t = np.arcsin((n1 / n2) * np.sin(theta))

    Rs = (
            np.abs(
                (n1 * np.cos(theta) - n2 * np.cos(theta_t))
                / (n1 * np.cos(theta) + n2 * np.cos(theta_t))
            )
            ** 2
    )


    Rp = (
            np.abs(
                (n1 * np.cos(theta_t) - n2 * np.cos(theta))
                / (n1 * np.cos(theta_t) + n2 * np.cos(theta))
            )
            ** 2
    )

    return Rs, Rp, np.array([0]), 1 - Rs, 1 - Rp

def calc_RAT_TMM(theta, pol, *args):
    lookuptable = args[0]

    angles = xr.DataArray(theta, dims=['unique_direction', 'wl_angle'])
    wls = xr.DataArray(lookuptable.wl.data, dims='wl_angle')

    data = lookuptable.sel(
        angle=angles, wl=wls, method="nearest"
    )

    # rearrange coordinates:
    Rs = data.R.sel(pol="s").transpose("unique_direction", "wl_angle")
    Rp = data.R.sel(pol="p").transpose("unique_direction", "wl_angle")

    Ts = data.T.sel(pol="s").transpose("unique_direction", "wl_angle")
    Tp = data.T.sel(pol="p").transpose("unique_direction", "wl_angle")

    A_per_layer = np.sum(data.Alayer.transpose("layer", "unique_direction", "wl_angle", "pol")*pol, -1)
    A_per_layer = A_per_layer.transpose("unique_direction", "layer", "wl_angle")
    return (np.real(Rs.data), np.real(Rp.data),
            np.real(A_per_layer.data),
            np.real(Ts.data), np.real(Tp.data))

class dummy_prop_rays:

    def __init__(self):
        # the only thing this needs to do is return None when used with .isel(wl=i)
        pass

    def isel(self, wl):
        return None

class zero_intensity_rays:

    def __init__(self):
        # the only thing this needs to do is return I=0 when used with .isel(wl=i)
        self.I = 0

    def isel(self, wl):
        return State(I = 0)

def update_absorbed_details(x, A, n_interactions, n_passes):

    if A.ndim == 1:
        A = A[None, :]

    if n_passes.ndim == 0:
        n_passes = np.array([n_passes]*A.shape[0])

    if n_interactions.ndim == 0:
        n_interactions = np.array([n_interactions]*A.shape[0])

    x.append(xr.Dataset({
                "A": xr.DataArray(A,
                                  dims=["unique_direction", "wl"]),
                "n_interactions": xr.DataArray(n_interactions,
                                               dims=["unique_direction"]),
                "n_passes": xr.DataArray(n_passes, dims=["unique_direction"]),
            }))


def analytical_start(nks,
                alphas,
                theta,
                r_a_0,
                phi,
                surfaces,
                widths,
                z_pos,
                depths,
                depth_indices,
                pol,
                initial_mat,
                initial_dir,
                tmm_args,
                max_interactions,
                wls
                     ):

    # if light is incident on a planar surface at an off-normal angle, different wavelengths will
    # have different angles of incidence! This should be fine but not currently expected by
    # analytical_per_facet

    # # generally same across wavelengths, but can be changed by analytical
    # # ray tracing happening first

    if initial_dir == -1:
        raise ValueError("Incidence from below (direction = -1) not implemented for analytical "
                         "ray-tracing")

    if np.sum(tmm_args[0]) == 0:
        # only Fresnel surfaces and tmm_args is just a list of zeroes.
        tmm_args = [0] + [len(surfaces)*[0]]
        # need tmm_args to have a second element, so that inner functions can correctly
        # identify that they need to use Fresnel and not TMM

    n_wl = len(wls)
    wls = xr.DataArray(wls, dims='wl_angle')

    current_pol = np.tile(pol, (n_wl, 1)) # "length"**2 of E-field vectors in s and p directions.
    # sum is 1.

    _, current_pol_vectors = make_pol_vectors(pol, theta, phi)

    # want the dimensions for current_pol_vectors to be: (pol, xyz, wl), need to repeat n_wl times
    # along new axis which will be the final index:
    current_pol_vectors = np.repeat(current_pol_vectors[:, :, None], n_wl, axis=2)

    mat_i = initial_mat

    # if initial_dir == 1: # travelling down
    surf_index = initial_mat
    # next_mat = initial_mat + initial_dir

    # else: # travelling up
    #     surf_index = initial_mat - 1

    profile = np.zeros((len(z_pos), n_wl))
    # do everything in microns
    A_per_layer = np.zeros((len(widths), n_wl))

    A_per_interface = [[] for _ in range(len(surfaces))]

    # make xarrays with attribute I = 0:
    overall_R = xr.Dataset({"I": xr.DataArray(np.zeros((1, n_wl)),
                                              dims=["unique_direction", "wl"])})
    overall_T = xr.Dataset({"I": xr.DataArray(np.zeros((1, n_wl)),
                                              dims=["unique_direction", "wl"])})

    theta = theta*np.ones(n_wl)
    angles = xr.DataArray(theta, dims='wl_angle')

    single_direction = True

    d = -r_a_0 / np.linalg.norm(r_a_0)

    I_remaining = xr.DataArray(np.ones((1, n_wl)), dims=["unique_direction", "wl"])

    n_interactions = 0
    n_passes = 0
    a_details = []
    prop_rays = []

    while single_direction:

        I_rem_data = I_remaining.data[0]

        R_data, A_data, T_data, R_pol, T_pol, A_int_detail = analytical_per_face(surfaces[surf_index],
                                      surf_index,
                                      d,
                                      tmm_args,
                                      nks,
                                      initial_dir,
                                      current_pol,
                                      current_pol_vectors,
                                      max_interactions[surf_index],
                                      )

        # Note: the polarization vector direction will not be updated here. In current
        # implementation with rt_structure, this is ok, because the calculation ends here.
        # If continuing the calculation, will need to calculate s and p vectors from the
        # ray direction; do need to store the current_pol


        A_per_interface[surf_index] = A_data*I_rem_data[None, None, :]

        R_data['n_interactions'] = R_data["n_interactions"] + n_interactions
        T_data['n_interactions'] = T_data["n_interactions"] + n_interactions

        n_int_detail = np.arange(1, A_int_detail.shape[0] + 1) + n_interactions

        A_int_detail = A_int_detail*I_rem_data[None, :]

        for i1, Ada in enumerate(A_int_detail):
            update_absorbed_details(a_details, Ada,   np.array([n_int_detail[i1]]), np.array([n_passes]))

        # scale R_data and T_data by I_remaining:
        R_data['I'] = R_data.I * I_rem_data[None, :]
        T_data['I'] = T_data.I * I_rem_data[None, :]

        if mat_i == initial_mat:
            n_passes_R = 0

        else:
            n_passes_R = n_passes + 1

        n_passes += 1

        R_data = xr.merge([R_data, xr.DataArray(n_passes_R).rename("n_passes")])
        T_data = xr.merge([T_data, xr.DataArray(n_passes).rename("n_passes")])

        current_pol = T_pol

        # surf_index only right for incidence from above
        DA, I = traverse_vectorised(
            widths[surf_index + initial_dir], # units?
            T_data.theta_t.data,
            alphas[surf_index + initial_dir], # units?
            np.ones_like(T_data.theta_t),
            depths[surf_index + initial_dir],
            initial_dir,
        )

        if surf_index == 0 and initial_dir == 1:
            # any rays that were reflected here are reflected overall into the incidence medium
            overall_R = R_data
            # do not want to include reflected rays in propagating rays, they are
            # accounted for in overall_R

        else:
            # could have multiple planar surfaces (though should really just make them
            # all part of the same surface in that case, with incoherent layers if necessary!),
            # and in that case we want to record rays travelling upwards here but continue with
            # analytical ray tracing

            # need to propagate these rays through the bulk and account for attenuation
            # of these rays and absorption in the bulk:
            theta_R = np.arccos(
                R_data.direction[:, 2] / np.linalg.norm(R_data.direction, axis=1))

            DA_R, I_R = traverse_vectorised(
                widths[mat_i],  # units?
                theta_R.data,
                alphas[mat_i],  # units?
                np.ones_like(theta_R.data),
                depths[mat_i],
                initial_dir,
            )

            I_abs_R = 1 - I_R

            I_out_per_direction_R = R_data.I.data * I_abs_R
            update_absorbed_details(a_details, I_out_per_direction_R, R_data.n_interactions, R_data.n_passes)

            A_actual_R = np.sum(I_out_per_direction_R, axis=0)
            DA_actual_R = np.sum(R_data.I.data.T * DA_R, axis=2)

            A_per_layer[mat_i] += np.real(A_actual_R)
            profile[depth_indices[mat_i]] += np.real(+ DA_actual_R)

            R_remaining = R_data.I * I_R
            prop_rays.append(xr.Dataset(
                {
                    "I": R_remaining,
                    "direction": R_data.direction,
                    "mat_i": mat_i,
                    "n_interactions": R_data.n_interactions,
                    "n_passes": R_data.n_passes,
                    "pol": xr.DataArray(R_pol, dims=["unique_direction", "wl", "sp"]),
                }
            )
            )

        if surf_index == len(surfaces) - 1 and initial_dir == 1:
            # any rays that were transmitted here are transmitted overall into the transmission medium
            overall_T = T_data
            include_T = False  # do not want to include transmitted rays in propagating rays, they are
            # accounted for in overall_T
            # TODO: does this need to be scaled?

        else:
            include_T = True

        I_abs = 1 - I

        I_out_per_direction = T_data.I.data * I_abs
        update_absorbed_details(a_details, I_out_per_direction, T_data.n_interactions,
                                            T_data.n_passes)
        A_actual = np.sum(I_out_per_direction, axis=0)

        DA_actual = np.sum(T_data.I.data.T*DA, axis=2)

        surf_index += initial_dir
        mat_i += initial_dir

        A_per_layer[mat_i] += np.real(A_actual)
        profile[depth_indices[mat_i]] += np.real(+ DA_actual)
        remaining_after_bulk = np.real(T_data.I * I)

        in_structure = surf_index < len(surfaces) and surf_index >= 0

        # if all rays are still travelling in the same direction, continue with analytical RT. Otherwise continue on to
        # 'normal' ray tracing. Otherwise, or if we have check the last surface in the structure,
        # end and return results.

        if (np.unique(T_data.direction, axis=0).shape[0] > 1 or
                not in_structure or surfaces[surf_index-1].phong):

            # single_direction = False
            # end, need to save/return final results here
            # absorption (profile and total) in bulk have been tracked as we went along
            # Need to save results for of remaining intensities and directions for each wavelength,
            # and where these rays are - note that we have already accounted for traversal of the
            # bulk layer, so these rays should start right before the next surface in 'normal'
            # ray-tracing.

            # Also need to save overall reflection and transmission into semi-infinite surrounding
            # which has happened so far (this should have been done above)

            # array with dimensions: (face, wl)
            # dataarrays for: direction (xyz), intensity, number of interactions,

            if include_T:

                if T_pol.ndim == 2:
                    T_pol = T_pol[None, :, :]

                prop_rays.append(xr.Dataset(
                    {
                        "I": remaining_after_bulk,
                        "direction": T_data.direction,
                        "mat_i": mat_i,
                        "n_interactions": T_data.n_interactions,
                        "n_passes": T_data.n_passes,
                        "pol": xr.DataArray(T_pol, dims=["unique_direction", "wl", "sp"]),
                    }
                )
                )

            # stack prop_rays along the unique_direction axis:
            if len(prop_rays) > 0:
                prop_rays = xr.concat(prop_rays, dim="unique_direction")

            else:
                prop_rays = zero_intensity_rays()

            a_details = xr.concat(a_details, dim="unique_direction")
            a_details["A"] = np.real(a_details["A"])

            # set nan values to 0:
            a_details["A"] = a_details["A"].fillna(0)

            R_to_add = np.sum(overall_R.I, axis=0).data
            T_to_add = np.sum(overall_T.I, axis=0).data

            A_interface_to_add = []
            for darray in A_per_interface:
                if np.sum(darray) > 0:
                    A_interface_to_add.append(np.sum(darray.data, axis=0).T)

                else:
                    A_interface_to_add.append(0)

            # return profile.T, A_per_layer.T, a_details, A_per_interface, overall_R, overall_T, prop_rays
            result_per_wl = {"R": R_to_add, "T": T_to_add, "A_per_layer": A_per_layer.T,
                             "A_per_interface": A_interface_to_add, "profile": profile.T}
            result_detail = {"overall_R": overall_R, "overall_T": overall_T, "A_details": a_details}

            return prop_rays, result_per_wl, result_detail

        else:
            # continue, but need to update inputs: transmitted rays at each wavelength become new
            # incident rays.
            angles.data = T_data.theta_t.data[0]
            theta = T_data.theta_t.data[0]
            I_remaining = remaining_after_bulk

            # can only reach here if surfaces so far have been planar; end up with
            # two directions because planar surface is made of two triangles, but they
            # contain the same information
            # s vector does not need to be changed if previous surface was planar
            I_remaining = I_remaining.sum(dim='unique_direction').expand_dims('unique_direction')

            d = T_data.direction[0].expand_dims('unique_direction').data
            new_p_vector = [np.cos(np.real(angles.data)), np.zeros_like(np.real(angles.data)), -np.sin(np.real(angles.data))]
            current_pol_vectors[1] = new_p_vector
            # current_pol = current_pol # if analytical_planar only
            current_pol =  current_pol[0]


def analytical_per_face(current_surf,
                         surf_index,
                         r_in,
                         tmm_args,
                         nks,
                         direction,
                         current_pol,
                         current_pol_vectors,
                         max_interactions,
                        ):

    n_wavelengths = nks.shape[1]
    how_many_faces = len(current_surf.N)
    normals = current_surf.N

    if current_pol.ndim == 2:
        current_pol = current_pol[None, :, :]

    if current_pol_vectors.ndim == 3:
        current_pol_vectors = np.tile(current_pol_vectors, (how_many_faces, 1, 1, 1))
        current_pol_vectors = np.moveaxis(current_pol_vectors, 0, 1)
        # indexing is (pol, face, xyz, wl)

    if tmm_args[0] > 0:
        n_layers = tmm_args[4][surf_index]

    else:
        n_layers = 0

    # TODO: only correct for downwards
    n0 = nks[surf_index]
    n1 = nks[surf_index + direction]

    opposite_faces = np.where(np.dot(normals, normals.T) < 0)[1]

    if tmm_args[1][surf_index] == 0:
        calc_RAT = calc_RAT_Fresnel
        R_args = [n0, n1]
        # TODO: above only correct for downwards

    else:
        calc_RAT = calc_RAT_TMM
        structpath = tmm_args[2]
        surf_name = tmm_args[3][surf_index] + "int_{}".format(surf_index)
        lookuptable = xr.open_dataset(os.path.join(structpath, surf_name + ".nc")).loc[dict(pol=['s', 'p'], side=direction)]
        # do I want to load this?
        R_args = [lookuptable]

    if len(r_in.flatten()) == 3:

        r_inc = np.tile(r_in[:,None], (how_many_faces, 1, n_wavelengths))  # (4, 3) array
        # r_inc = r_inc[:, :, None]

    else:
        # (4, 3, n_wavelengths array):
        r_inc = np.tile(r_in, (how_many_faces, 1, 1))


    area = np.sqrt(
        np.sum(np.cross(current_surf.P_0s - current_surf.P_1s, current_surf.P_2s - current_surf.P_1s, axis=1) ** 2, 1)
        ) / 2

    relevant_face = np.arange(how_many_faces)

    R_per_it = np.zeros((how_many_faces, max_interactions, n_wavelengths))
    T_per_it = np.zeros((how_many_faces, max_interactions, n_wavelengths))
    T_dir_per_it = np.zeros((how_many_faces, max_interactions, n_wavelengths))
    T_pol_per_it = np.zeros((how_many_faces, max_interactions, n_wavelengths, 2))
    A_per_it = np.zeros((how_many_faces, n_layers, max_interactions, n_wavelengths))

    stop_it = np.ones(how_many_faces, dtype=int) * max_interactions

    cos_inc = -np.sum(normals[relevant_face, :, None] * r_inc, 1)  # dot product

    hit_prob = area[relevant_face, None] * cos_inc  # scale by area of each triangle
    hit_prob[
        cos_inc < 0] = 0  # if negative, then the ray is shaded from that pyramid face and will never hit it
    hit_prob = hit_prob / np.sum(hit_prob, axis=0)  # initial probability of hitting each face

    reflected_ray_directions = np.zeros((how_many_faces, 3, max_interactions, n_wavelengths))
    transmitted_ray_directions = np.zeros((how_many_faces, 3, max_interactions, n_wavelengths))

    N_interaction = 0

    while np.sum(cos_inc) > 0:

        reflected_direction = r_inc - 2 * np.sum(r_inc*normals[relevant_face, :, None], axis=1)[:, None] * normals[relevant_face, :, None]
        reflected_direction = reflected_direction / np.linalg.norm(reflected_direction, axis=1)[:, None]

        reflected_ray_directions[:, :, N_interaction] = reflected_direction

        # if negative, then the ray is shaded from that pyramid face and will never hit it

        tr_par = (n0 / n1) * (r_inc - np.sum(r_inc * normals[relevant_face, :, None], axis=1)[:,
                                      None] * normals[relevant_face, :, None])
        tr_par_norm = np.linalg.norm(tr_par, axis=1)
        tr_par_norm[tr_par_norm > 1] = 1
        tr_perp = -np.sqrt(1 - tr_par_norm ** 2)[:, None, :] * normals[relevant_face, :, None]

        refracted_rays = np.real(tr_par + tr_perp)
        refracted_rays  = refracted_rays / np.linalg.norm(refracted_rays, axis=1)[:,None, :]
        transmitted_ray_directions[:, :,  N_interaction] = refracted_rays

        normals_relevant = normals[relevant_face][:, :, None]

        ray_plane_s_direction = np.cross(r_inc, normals_relevant, axis=1)

        # this gives 0s if the ray is parallel to the normal
        norm_s_dir = np.linalg.norm(ray_plane_s_direction, axis=1)

        # find indices of places where this happens:
        is_parallel = np.abs(norm_s_dir) < 1e-10
        # this has dimensions (face, wl)

        if np.any(is_parallel):
            ray_plane_s_temp = np.moveaxis(ray_plane_s_direction, 1, -1)
            current_s_vector = np.moveaxis(current_pol_vectors[0], 1, -1)
            # if the ray is parallel to the normal, then the s direction is the same as the
            # previous s direction

            ray_plane_s_temp[is_parallel] = current_s_vector[is_parallel]

            ray_plane_s_direction = np.moveaxis(ray_plane_s_temp, -1, 1)


        ray_plane_s_direction = ray_plane_s_direction / np.linalg.norm(ray_plane_s_direction,
                                                                       axis=1)[:,None,:]

        s_s = np.einsum('ijk,ijk->ik', current_pol_vectors[0], ray_plane_s_direction)
        p_s = np.einsum('ijk,ijk->ik', current_pol_vectors[1], ray_plane_s_direction)
        # dimensions: (faces, n_wl)

        s_component_sq = (current_pol[:, :, 0] * s_s ** 2 + current_pol[:, :, 1] * p_s ** 2)
        p_component_sq = 1 - s_component_sq

        pol_weight = np.moveaxis(np.array([s_component_sq, p_component_sq]), 0, -1)
        Rs_prob, Rp_prob, A_prob, Ts_prob, Tp_prob = calc_RAT(np.arccos(cos_inc), pol_weight, *R_args)
        # these are for light which is s and p polarized with respect to each face; need to
        # find components of each ray which are s and p polarized with respect to the
        # existing ray's s and p directions
        reflected_p_direction = np.cross(r_inc, ray_plane_s_direction, axis=1)

        current_pol_vectors = np.stack([ray_plane_s_direction, reflected_p_direction])
        # dimensions of Rs_prob etc.: (n_faces, n_wls)
        # dimensions of normals[relevant_face]: (n_faces, 3)
        # dimensions of current_pol_vectors: (sp, xyz, n_wl)
        # dimensions of r_inc: (n_faces, 3, n_wls)

        R_stack = np.stack((Rs_prob*s_component_sq, Rp_prob*p_component_sq), axis=-1)
        R_prob = np.sum(R_stack, -1)

        # stack Ts_prob and Tp_prob so that the final index of the array is the new one:
        T_stack = np.stack((Ts_prob*s_component_sq, Tp_prob*p_component_sq), axis=-1)

        # fixed (? not checked) up to here for vector polarization directions
        current_pol = R_stack/R_prob[:,:,None]
        # current_pol = current_pol / (np.sum(current_pol, -1)[:, :, None])

        if np.sum(A_prob) > 0:
            A_prob_sum = np.sum(A_prob, axis=1)

        else:
            A_prob_sum = 0

        T_prob = 1 - R_prob - A_prob_sum

        T_per_it[:, N_interaction] = T_prob

        A_per_it[:, :, N_interaction] = A_prob

        T_dir_per_it[:, N_interaction] = np.abs(
            refracted_rays[:, 2] / np.linalg.norm(refracted_rays,
                                                  axis=1))  # cos (global) of refracted ray

        T_pol_per_it[:, N_interaction] = T_stack/T_prob[:,:,None]
        cos_inc[reflected_direction[:, 2] > 0] = 0
        stop_it[
            np.all((np.all(reflected_direction[:, 2] > 0, axis=1), stop_it > N_interaction),
                   axis=0)] = N_interaction
         # want to end for this surface, since rays are travelling upwards -> no intersection

        R_per_it[:,N_interaction] = R_prob  # intensity reflected from each face, relative to incident total intensity 1

        # once ray travels upwards once, want to end calculation for that plane; don't want to
        # double count

        if len(opposite_faces) > 0:
            relevant_face = opposite_faces[relevant_face]

        r_inc = reflected_direction

        if np.sum(cos_inc) == 0:
            # no more interactions with any of the faces
            break

        cos_inc = -np.sum(normals[relevant_face, :, None] * r_inc, 1)  # dot product
        cos_inc[cos_inc < 0] = 0

        N_interaction += 1

    remaining_intensity = np.insert(np.cumprod(R_per_it, axis=1), 0, np.ones((how_many_faces, n_wavelengths)),
                                    axis=1)[:, :-1]

    R_total = np.array([hit_prob[j1] * np.prod(R_per_it[j1, :stop_it[j1] + 1], axis=0) for j1 in
               range(how_many_faces)])
    final_R_directions = np.array([reflected_ray_directions[j1, :, stop_it[j1]] for j1 in
                          range(how_many_faces)])
    final_R_pol = current_pol

    # the weight of each of these directions is R_total

    # loop through faces and interactions:
    final_T_directions = []
    final_T_weights = []
    final_T_n_interactions = []
    final_T_pol = []

    for j1 in range(how_many_faces):
        for j2 in range(stop_it[j1] + 1):
            final_T_directions.append(transmitted_ray_directions[j1, :, j2])
            final_T_weights.append(hit_prob[j1]*remaining_intensity[j1, j2]*T_per_it[j1, j2])
            final_T_n_interactions.append(j2 + 1)
            final_T_pol.append(T_pol_per_it[j1, j2])

    final_T_weights = np.array(final_T_weights) # is this a function of wavelength?
    final_T_weights[final_T_weights < 0] = 0
    final_T_directions = np.array(final_T_directions)
    final_T_pol = np.array(final_T_pol)
    final_T_pol = final_T_pol / np.sum(final_T_pol, -1)[:, :, None]

    # A_per_it has dimensions: (face, layer, interaction, wavelength)
    # sum over layers:
    A_per_it_sum = np.sum(hit_prob[:,None]*remaining_intensity*np.sum(A_per_it, axis=1), 0)
    # A_per_it_sum is the total absorption in the interface (all faces) per interaction

    A_total = hit_prob[:, None] * np.sum(remaining_intensity[:, None, :, :] * A_per_it, axis=2)

    # theta_out_R = np.arccos(final_R_directions[:, 2] / np.linalg.norm(final_R_directions, axis=1))
    # phi_out_R = np.arctan2(final_R_directions[:, 1], final_R_directions[:, 0])
    # number of reps of each theta value for the angular distribution:

    theta_out_T = np.arccos(final_T_directions[:, 2] / np.linalg.norm(final_T_directions, axis=1))
    # phi_out_T = np.arctan2(final_T_directions[:, 1], final_T_directions[:, 0])

    # list of results and their dimensions:
    # R_total: (face, wavelength)
    # final_R_directions: (face, 3, wavelength)
    # theta_out_R: (face, wavelength)
    # phi_out_R: (face, wavelength)

    # A_total: (face, layer, wavelength)

    # final_T_weights: (number of outgoing directions, wavelength)
    # final_T_directions: (number of outgoing directions, 3, wavelength)
    # final_T_n_interactions: (number of outgoing directions)
    # theta_out_T: (number of outgoing directions, wavelength)
    # phi_out_T: (number of outgoing directions, wavelength)

    # make xarrays for each of these:
    total_I = np.sum(R_total, 0) + np.sum(final_T_weights, 0) + np.sum(np.sum(A_total, 1), 0)

    if not np.allclose(total_I, 1, rtol=5e-3, atol=5e-3):
        logger.warning(f"Total intensity not conserved in analytical ray tracing, indicating "
                      f"the number of interactions is not constant across the unit cell. Maximum "
                      f"deviation: {100*np.max(np.abs(total_I - 1)):.1f} %")

    R_total = xr.DataArray(R_total, dims=["unique_direction", "wl"])
    final_R_directions = xr.DataArray(final_R_directions, dims=["unique_direction", "xyz", "wl"])
    n_interactions = xr.DataArray(stop_it + 1, dims=["unique_direction"])
    # theta_out_R = xr.DataArray(theta_out_R, dims=["unique_direction"])
    # phi_out_R = xr.DataArray(phi_out_R, dims=["unique_direction"])

    # if all rays originating from a surface had the same number of interactions, this should sum
    # to:

    R_data = xr.Dataset(
        {
            "I": R_total/total_I,
            "direction": final_R_directions,
            "n_interactions": n_interactions,
        }
    )

    A_data = xr.DataArray(A_total/total_I[None, None, :], dims=["unique_direction", "layer", "wl"])

    final_T_weights = xr.DataArray(final_T_weights, dims=["unique_direction", "wl"])
    final_T_directions = xr.DataArray(final_T_directions, dims=["unique_direction", "xyz", "wl"])
    final_T_n_interactions = xr.DataArray(final_T_n_interactions, dims=["unique_direction"])
    theta_out_T = xr.DataArray(theta_out_T, dims=["unique_direction", "wl"])

    T_data = xr.Dataset(
        {
            "I": final_T_weights/total_I,
            "direction": final_T_directions,
            "n_interactions": final_T_n_interactions,
            "theta_t": theta_out_T,
        }
    )

    return R_data, A_data, T_data, final_R_pol, final_T_pol, A_per_it_sum



# Copyright (C) 2021-2024 Phoebe Pearce
#
# This file is part of RayFlare and is released under the GNU Lesser General Public License (LGPL), version 3.
# Please see the LICENSE.txt file included as part of this package.
#
# Contact: p.pearce@unsw.edu.au

import numpy as np
import os
from cmath import sin, cos
from itertools import product
import xarray as xr
from sparse import COO, save_npz, stack
from joblib import Parallel, delayed
from warnings import warn

from rayflare.angles import fold_phi, make_angle_vector, overall_bin
from rayflare.utilities import get_matrices_or_paths
from .rt_common import Ray, single_interface_check, make_pol_vectors

from rayflare import logger


def RT(
    group,
    incidence,
    transmission,
    surf_name,
    options,
    structpath,
    Fr_or_TMM=0,
    front_or_rear="front",
    n_absorbing_layers=0,
    calc_profile=None,
    only_incidence_angle=False,
    widths=None,
    save=True,
    overwrite=False,
    analytical=False,
):
    """Calculates the reflection/transmission and absorption redistribution matrices for an interface using
    either a previously calculated TMM lookup table or the Fresnel equations.

    :param group: an RTgroup object containing the surface textures
    :param incidence: incidence medium
    :param transmission: transmission medium
    :param surf_name: name of the surface (to save matrices)
    :param options: user options (State object)
    :param structpath: file path where matrices will be stored or loaded from
    :param Fr_or_TMM: whether to use the Fresnel equations (0) or a TMM lookup table (1)
    :param front_or_rear: whether light is incident from the front or rear
    :param n_absorbing_layers: for a structure with multiple interface layers, where a TMM lookuptable is used, the number of layers in \
    the interface
    :param calc_profile: whether to save the relevant information to calculate the depth-dependent absorption \
    profile. List of layers where the profile should be calculated, or otherwise None
    :param only_incidence_angle: if True, the ray-tracing will only be performed for the incidence theta and phi \
    specified in the options.
    :param widths: if using TMM, width of the surface layers (in nm)
    :param save: whether to save redistribution matrices (True/False), default True
    :param overwrite: whether to overwrite existing matrices (True/False), default False

    :return: Number of returns depends on whether absorption profiles are being calculated; the first two items are
             always returned, the final one only if a profile is being calcualted.

                - allArrays: the R/T redistribution matrix at each wavelength, indexed as (wavelength, angle_bin_out, angle_bin_in)
                - absArrays: the absorption redistribution matrix (total absorption per layer), indexed as (wavelength, layer_out, angle_bin_in)
                - allres: xarray dataset storing the absorption profile data
    """

    existing_mats, path_or_mats = get_matrices_or_paths(
        structpath, surf_name, front_or_rear, calc_profile, overwrite
    )

    if existing_mats and not overwrite:
        return path_or_mats

    else:
        wavelengths = options.wavelength
        n_rays = options.n_rays
        nx = options.nx
        ny = options.ny
        n_angles = int(np.ceil(n_rays / (nx * ny)))

        phi_sym = options.phi_symmetry
        n_theta_bins = options.n_theta_bins
        c_az = options.c_azimuth

        pol = options.pol

        # if not (pol[0] == 1 or pol[1] == 1):
        #     logger.warning("Warning: you have specificied unpolarized/partially polarized light. "
        #                    "The ARRM RT class does not currently take into account polarization "
        #                    "changes at interfaces.")

        if not options.parallel:
            n_jobs = 1

        else:
            n_jobs = options.n_jobs if "n_jobs" in options else -1

        if calc_profile is not None:
            depth_spacing = options.depth_spacing * 1e9  # convert from m to nm
        else:
            depth_spacing = None

        if front_or_rear == "front":
            side = 1
        else:
            side = -1

        if Fr_or_TMM == 1:
            lookuptable = xr.open_dataset(os.path.join(structpath, surf_name + ".nc"))
            if front_or_rear == "rear":
                # side gets flipped here
                lookuptable = lookuptable.assign_coords(side=np.flip(lookuptable.side))
        else:
            lookuptable = None

        theta_spacing = options.theta_spacing if "theta_spacing" in options else "sin"

        theta_intv, phi_intv, angle_vector = make_angle_vector(
            n_theta_bins, phi_sym, c_az, theta_spacing
        )

        if only_incidence_angle:
            logger.info("Calculating matrix only for incidence theta/phi")
            if options.theta_in == 0:
                th_in = 0.0001
            else:
                th_in = options.theta_in

            angles_in = angle_vector[: int(len(angle_vector) / 2), :]
            n_reps = int(np.ceil(n_angles / len(angles_in)))
            thetas_in = np.tile(th_in, n_reps)
            n_angles = n_reps

            if options.phi_in == "all":
                # get relevant phis
                phis_in = np.tile(options.phi_in, n_reps)
            else:
                if options.phi_in == 0:
                    phis_in = np.tile(0.0001, n_reps)

                else:
                    phis_in = np.tile(options.phi_in, n_reps)

        else:
            if options.random_ray_angles:
                thetas_in = np.random.random(n_angles) * np.pi / 2
                phis_in = np.random.random(n_angles) * 2 * np.pi
            else:
                angles_in = angle_vector[: int(len(angle_vector) / 2), :]
                if n_angles / len(angles_in) < 1:
                    warn(
                        "The number of rays is not sufficient to populate the redistribution matrix!"
                    )
                n_reps = int(np.ceil(n_angles / len(angles_in)))
                thetas_in = np.tile(angles_in[:, 1], n_reps)[:n_angles]
                phis_in = np.tile(angles_in[:, 2], n_reps)[:n_angles]

        if front_or_rear == "front":
            mats = [incidence]
        else:
            mats = [transmission]

        if group.materials is not None:
            for mat_i in group.materials:
                mats.append(mat_i)

        if front_or_rear == "front":
            mats.append(transmission)
        else:
            mats.append(incidence)

        # list of lists: first in tuple is front incidence
        if front_or_rear == "front":
            surfaces = [x[0] for x in group.textures]

        else:
            surfaces = [x[1] for x in group.textures]

        nks = np.empty((len(mats), len(wavelengths)), dtype=complex)

        for i1, mat in enumerate(mats):
            nks[i1] = mat.n(wavelengths) + 1j * mat.k(wavelengths)

        h = max(surfaces[0].Points[:, 2])
        x_limits = (
            options.x_limits
            if "x_limits" in options
            else [
                surfaces[0].x_min + 0.01 * surfaces[0].Lx,
                surfaces[0].x_max - 0.01 * surfaces[0].Lx,
            ]
        )
        y_limits = (
            options.y_limits
            if "y_limits" in options
            else [
                surfaces[0].y_min + 0.01 * surfaces[0].Ly,
                surfaces[0].y_max - 0.01 * surfaces[0].Ly,
            ]
        )

        if options.random_ray_position:
            xs = np.random.uniform(x_limits[0], x_limits[1], nx)
            ys = np.random.uniform(y_limits[0], y_limits[1], ny)

        else:
            xs = np.linspace(x_limits[0], x_limits[1], nx)
            ys = np.linspace(y_limits[0], y_limits[1], ny)

        if analytical:
            pass

        else:

            allres = Parallel(n_jobs=n_jobs)(
                delayed(RT_wl)(
                    i1,
                    wavelengths[i1],
                    n_angles,
                    nx,
                    ny,
                    widths,
                    thetas_in,
                    phis_in,
                    h,
                    xs,
                    ys,
                    nks,
                    surfaces,
                    pol,
                    (np.pi / 2) / options.lookuptable_angles,
                    phi_sym,
                    theta_intv,
                    phi_intv,
                    angle_vector,
                    Fr_or_TMM,
                    n_absorbing_layers,
                    lookuptable,
                    calc_profile,
                    depth_spacing,
                    side,
                )
                for i1 in range(len(wavelengths))
            )

            allArrays = stack([item[0] for item in allres])
            absArrays = stack([item[1] for item in allres])

        if save:
            save_npz(path_or_mats[0], allArrays)
            save_npz(path_or_mats[1], absArrays)

        if Fr_or_TMM > 0 and calc_profile is not None:
            profile = xr.concat([item[3] for item in allres], "wl")
            intgr = xr.concat([item[4] for item in allres], "wl")
            intgr.name = "intgr"
            profile.name = "profile"

            intgr = intgr.where(intgr > 0, 0)
            profile = profile.where(profile > 0, 0)

            allres = xr.merge([intgr, profile])

            if save:
                allres.to_netcdf(path_or_mats[2])

            return allArrays, absArrays, allres

        else:
            return allArrays, absArrays


def RT_wl(
    i1,
    wl,
    n_angles,
    nx,
    ny,
    widths,
    thetas_in,
    phis_in,
    h,
    xs,
    ys,
    nks,
    surfaces,
    pol,
    d_theta,
    phi_sym,
    theta_intv,
    phi_intv,
    angle_vector,
    Fr_or_TMM,
    n_abs_layers,
    lookuptable,
    calc_profile,
    depth_spacing,
    side,
):

    if lookuptable is not None:
        lookuptable_wl_sp = lookuptable.sel(pol=["s", "p"]).sel(wl=wl * 1e9).load()
        lookuptable_wl = lookuptable.sel(wl=wl * 1e9).load()

    else:
        lookuptable_wl = None
        lookuptable_wl_sp = None

    logger.info(f"RT calculation for wavelength = {wl * 1e9} nm")

    theta_out = np.zeros((n_angles, nx * ny))
    phi_out = np.zeros((n_angles, nx * ny))
    A_surface_layers = np.zeros((n_angles, nx * ny, n_abs_layers))
    theta_local_incidence = np.zeros((n_angles, nx * ny))
    pol_local_incidence = np.zeros((n_angles, nx * ny, 2))

    for i2 in range(n_angles):

        theta = thetas_in[i2]
        phi = phis_in[i2]
        r = abs((h + 1e-8) / cos(theta))
        r_a_0 = np.real(
            np.array(
                [r * sin(theta) * cos(phi), r * sin(theta) * sin(phi), r * cos(theta)]
            )
        )
        pol, pol_vectors = make_pol_vectors(pol, theta, phi)

        for c, vals in enumerate(product(xs, ys)):
            ray, th_o, phi_o, surface_A = single_ray_interface(
                vals[0],
                vals[1],
                nks[:, i1],
                r_a_0,
                theta,
                phi,
                surfaces,
                pol,
                pol_vectors,
                d_theta,
                wl,
                Fr_or_TMM,
                lookuptable_wl_sp,
            )

            if th_o < 0:  # can do outside loup with np.where
                th_o = -th_o
                phi_o = phi_o + np.pi
            theta_out[i2, c] = th_o
            phi_out[i2, c] = phi_o
            A_surface_layers[i2, c] = surface_A[0]
            theta_local_incidence[i2, c] = np.real(surface_A[1])
            pol_local_incidence[i2, c] = ray.pol

    phi_out = fold_phi(phi_out, phi_sym)
    phis_in = fold_phi(phis_in, phi_sym)

    if side == -1:
        not_absorbed = np.where(theta_out < (np.pi + 0.1))
        thetas_in = np.pi - thetas_in
        # phis_in = np.pi-phis_in # unsure about this part

        theta_out[not_absorbed] = np.pi - theta_out[not_absorbed]
        # phi_out = np.pi-phi_out # unsure about this part

    theta_local_incidence = np.abs(theta_local_incidence)
    n_thetas = len(theta_intv) - 1

    if Fr_or_TMM > 0:
        # now we need to make bins for the absorption
        theta_intv = np.append(theta_intv, 11)
        phi_intv = phi_intv + [np.array([0])]

    # xarray: can use coordinates in calculations using apply!
    binned_theta_in = np.digitize(thetas_in, theta_intv, right=True) - 1

    binned_theta_out = np.digitize(theta_out, theta_intv, right=True) - 1
    # -1 to give the correct index for the bins in phi_intv

    phi_in = xr.DataArray(
        phis_in,
        coords={"theta_bin": (["angle_in"], binned_theta_in)},
        dims=["angle_in"],
    )

    bin_in = (
        phi_in.groupby("theta_bin")
        .map(overall_bin, args=(phi_intv, angle_vector[:, 0]))
        .data
    )

    phi_out = xr.DataArray(
        phi_out,
        coords={"theta_bin": (["angle_in", "position"], binned_theta_out)},
        dims=["angle_in", "position"],
    )

    bin_out = (
        phi_out.groupby("theta_bin")
        .map(overall_bin, args=(phi_intv, angle_vector[:, 0]))
        .data
    )

    out_mat = np.zeros((len(angle_vector), int(len(angle_vector) / 2)))
    # everything is coming in from above so we don't need 90 -> 180 in incoming bins
    A_mat = np.zeros((n_abs_layers, int(len(angle_vector) / 2)))

    n_rays_in_bin = np.zeros(int(len(angle_vector) / 2))
    n_rays_in_bin_abs = np.zeros(int(len(angle_vector) / 2))

    binned_local_angles = np.digitize(theta_local_incidence, theta_intv, right=True) - 1

    local_angle_mat = np.zeros(
        (int((len(theta_intv) - 1) / 2), int(len(angle_vector) / 2))
    )

    if side == 1:
        offset = 0
    else:
        offset = int(len(angle_vector) / 2)

    for l1 in range(len(thetas_in)):
        for l2 in range(nx * ny):
            n_rays_in_bin[bin_in[l1] - offset] += 1
            if binned_theta_out[l1, l2] <= (n_thetas - 1):
                # reflected or transmitted
                out_mat[bin_out[l1, l2], bin_in[l1] - offset] += 1

            else:
                # absorbed in one of the surface layers
                n_rays_in_bin_abs[bin_in[l1] - offset] += 1
                per_layer = A_surface_layers[l1, l2]
                A_mat[:, bin_in[l1] - offset] += per_layer
                local_angle_mat[binned_local_angles[l1, l2], bin_in[l1] - offset] += 1

    # normalize
    out_mat = np.divide(out_mat, n_rays_in_bin, where=n_rays_in_bin != 0)
    overall_abs_frac = np.divide(
        n_rays_in_bin_abs, n_rays_in_bin, where=n_rays_in_bin != 0
    )
    abs_scale = np.divide(
        overall_abs_frac, np.sum(A_mat, 0), where=np.sum(A_mat, 0) != 0
    )

    intgr = np.divide(
        np.sum(A_mat, 0),
        n_rays_in_bin_abs,
        where=n_rays_in_bin_abs != 0,
        out=np.zeros_like(n_rays_in_bin_abs),
    )
    A_mat = abs_scale * A_mat
    out_mat[np.isnan(out_mat)] = 0
    A_mat[np.isnan(A_mat)] = 0

    # for some reason, sometimes get really small -ve values like -1e-300 instead of 0.
    # this causes issues and produces nans in the matrix multiplication (dot_wl) later, although
    # only on Ubuntu for some reason?
    out_mat[out_mat < 0] = 0
    A_mat[A_mat < 0] = 0

    out_mat = COO.from_numpy(out_mat)  # sparse matrix
    A_mat = COO.from_numpy(A_mat)

    # out_mat[out_mat < 0] = 0

    if Fr_or_TMM > 0:
        local_angle_mat = np.divide(
            local_angle_mat,
            np.sum(local_angle_mat, 0),
            where=np.sum(local_angle_mat, 0) != 0,
            out=np.zeros_like(local_angle_mat),
        )
        local_angle_mat[np.isnan(local_angle_mat)] = 0
        local_angle_mat = COO.from_numpy(local_angle_mat)

        if calc_profile is not None:
            n_a_in = int(len(angle_vector) / 2)
            thetas = angle_vector[:n_a_in, 1]
            unique_thetas = np.unique(thetas)

            profile = make_profiles_wl(
                unique_thetas,
                n_a_in,
                side,
                widths,
                local_angle_mat,
                lookuptable_wl,
                pol,
                depth_spacing,
                calc_profile,
            )

            profile = profile.rename({"dim_0": "z"})

            intgr = xr.DataArray(
                intgr,
                dims=["global_index"],
                coords={"global_index": np.arange(0, n_a_in)},
            ).fillna(0)

            return out_mat, A_mat, local_angle_mat, profile, intgr

        else:
            return out_mat, A_mat, local_angle_mat

    else:
        return out_mat, A_mat


def make_profiles_wl(
    unique_thetas,
    n_a_in,
    side,
    widths,
    angle_distmat,
    lookuptable,
    pol,
    depth_spacing,
    prof_layers,
):
    # widths and depth_spacing are passed in nm!

    def profile_per_layer(xx, z, offset, side, non_zero):
        layer_index = xx.coords["layer"].item(0) - 1
        x = xx.squeeze()
        x = x[non_zero]
        part1 = x[:, 0] * np.exp(x[:, 4] * z[layer_index])
        part2 = x[:, 1] * np.exp(-x[:, 4] * z[layer_index])
        part3 = (x[:, 2] + 1j * x[:, 3]) * np.exp(1j * x[:, 5] * z[layer_index])
        part4 = (x[:, 2] - 1j * x[:, 3]) * np.exp(-1j * x[:, 5] * z[layer_index])
        result = np.real(part1 + part2 + part3 + part4)
        if side == -1:
            result = np.flip(result, 1)
        return result.reduce(np.sum, axis=0).assign_coords(
            dim_0=z[layer_index] + offset[layer_index]
        )

    def profile_per_angle(x, z, offset, side, nz):
        i2 = x.coords["global_index"].item(0)
        non_zero = np.where(nz[:, i2])[0]
        by_layer = x.groupby("layer", squeeze=False).map(
            profile_per_layer, z=z, offset=offset, side=side, non_zero=non_zero
        )
        return by_layer

    def scale_func(x, scale_params):
        xx = x.squeeze()
        return xx.data[:, None, None] * scale_params

    def select_func(x, const_params):
        xx = x.squeeze()
        return (xx.data[:, None, None] != 0) * const_params

    pr = xr.DataArray(
        angle_distmat.todense(),
        dims=["local_theta", "global_index"],
        coords={"local_theta": unique_thetas, "global_index": np.arange(0, n_a_in)},
    )

    # lookuptable layers are 1-indexed
    if pol[0] > 0.999:
        pol_s = "s"

    elif pol[1] > 0.999:
        pol_s = "p"

    else:
        pol_s = "u"

    data = lookuptable.loc[dict(side=1, pol=pol_s)].interp(
        angle=pr.coords["local_theta"], #wl=wl * 1e9
    )

    params = (
        data["Aprof"]
        .drop_vars(["layer", "side", "angle", "pol"])
        .transpose("local_theta", "layer", "coeff")
    )

    s_params = params.loc[
        dict(coeff=["A1", "A2", "A3_r", "A3_i"])
    ]  # have to scale these to make sure integrated absorption is correct
    c_params = params.loc[dict(coeff=["a1", "a3"])]  # these should not be scaled

    scale_res = pr.groupby("global_index", squeeze=False).map(scale_func, scale_params=s_params)
    const_res = pr.groupby("global_index", squeeze=False).map(select_func, const_params=c_params)

    params = xr.concat((scale_res, const_res), dim="coeff").assign_coords(
        layer=np.arange(1, len(widths) + 1)
    )
    params = params.transpose("local_theta", "global_index", "layer", "coeff")

    z_list = []

    for l_w in widths:
        z_list.append(xr.DataArray(np.arange(0, l_w, depth_spacing)))

    offsets = np.cumsum([0] + widths)[:-1]

    xloc = params.loc[dict(coeff="A1")].reduce(np.sum, "layer")
    nz = xloc != 0

    ans = (
        params.loc[dict(layer=prof_layers)]
        .groupby("global_index")
        .map(profile_per_angle, z=z_list, offset=offsets, side=side, nz=nz)
        .drop_vars("coeff")
    )
    ans = ans.fillna(0)

    profile = ans.reduce(np.sum, "layer")
    profile = xr.where(profile >= 0, profile, 0)

    return profile.T


def single_ray_interface(
    x, y, nks, r_a_0, theta, phi, surfaces, pol, pol_vec, d_theta, wl, Fr_or_TMM, lookuptable
):
    direction = 1  # start travelling downwards; 1 = down, -1 = up
    mat_index = 0  # start in first medium
    surf_index = 0
    stop = False

    # could be done before to avoid recalculating every time
    r_a = r_a_0 + np.array([x, y, 0])
    r_b = np.array(
        [x, y, 0]
    )  # set r_a and r_b so that ray has correct angle & intersects with first surface
    d = (r_b - r_a) / np.linalg.norm(r_b - r_a)  # direction (unit vector) of ray

    ray = Ray(1, d, r_a, pol_vec[0], pol_vec[1], pol)

    while not stop:

        surf = surfaces[surf_index]

        ray.r_a[0] = ray.r_a[0] - surf.Lx * (
            (ray.r_a[0] + ray.d[0] * (surf.zcov - ray.r_a[2]) / ray.d[2]) // surf.Lx
        )
        ray.r_a[1] = ray.r_a[1] - surf.Ly * (
            (ray.r_a[1] + ray.d[1] * (surf.zcov - ray.r_a[2]) / ray.d[2]) // surf.Ly
        )

        res, theta, phi, theta_loc, _, _ = single_interface_check(
            ray,
            nks[mat_index],
            nks[mat_index + 1],
            surf,
            surf.Lx,
            surf.Ly,
            direction,
            surf.zcov,
            d_theta,
            0,
            wl,
            Fr_or_TMM,
            lookuptable,
        )

        if res == 0:  # reflection
            direction = -direction  # changing direction due to reflection

            # staying in the same material, so mat_index does not change, but surf_index does
            surf_index = surf_index + direction

            surface_A = [0, 10]

        if res == 1:  # transmission
            surf_index = surf_index + direction
            mat_index = mat_index + direction  # is this right?

            surface_A = [0, 10]

        if res == 2:
            surface_A = [
                theta,
                theta_loc,
            ]  # passed a list of absorption per layer in theta
            stop = True
            theta = 10  # theta returned by single_interface_check is actually list of absorption per layer

        if direction == 1 and mat_index == 1:

            stop = True
            # have ended with transmission

        elif direction == -1 and mat_index == 0:
            stop = True

    return ray, theta, phi, surface_A